from random import randrange

from pychess.System import conf
from pychess.System import uistuff
from pychess.widgets import mainwindow
from pychess.Savers.remotegame import get_internet_game_providers


class TipOfTheDay:
    def __init__(self):
        self.widgets = uistuff.GladeWidgets("tipoftheday.glade")
        self.widgets["window1"].set_transient_for(mainwindow())
        uistuff.keepWindowSize(
            "tipoftheday", self.widgets["window1"], (320, 240), uistuff.POSITION_CENTER
        )

        self.widgets["checkbutton1"].set_active(conf.get("show_tip_at_startup"))
        self.widgets["checkbutton1"].connect(
            "toggled", lambda w: conf.set("show_tip_at_startup", w.get_active())
        )
        self.widgets["close_button"].connect(
            "clicked", lambda w: self.widgets["window1"].emit("delete-event", None)
        )
        self.widgets["window1"].connect(
            "delete_event", lambda w, a: self.widgets["window1"].destroy()
        )
        self.widgets["back_button"].connect(
            "clicked", lambda w: self.set_currentIndex(self.tips_curindex - 1)
        )
        self.widgets["forward_button"].connect(
            "clicked", lambda w: self.set_currentIndex(self.tips_curindex + 1)
        )

        self.tips_fixed = 2
        self.tips = [
            # PyChess facts -- The first tips_fixed messages are always displayed first
            _(
                "PyChess is an open-source chess application that can be enhanced by any chess enthusiast: bug reports, source code, documentation, translations, feature requests, user assistance... Let's get in touch at <b>https://pychess.github.io</b>"
            ),
            _(
                "PyChess supports a wide range of chess engines, variants, Internet servers, and lessons. It is a perfect desktop application to conveniently improve your chess skills."
            ),
            _(
                "The releases of PyChess hold the name of historical world chess champions. Do you know the name of the current world chess champion?"
            ),
            _(
                "Did you know that you can help translate PyChess into your own language, <b>Help</b> > <b>Translate PyChess</b>."
            ),
            _(
                "A game consists of an opening, a middle-game and an end-game. PyChess is able to train you thanks to its opening book, its supported chess engines and its training module."
            ),
            # Chess facts
            _(
                "Did you know that it is possible to finish a chess game in just 2 turns?"
            ),
            _(
                "Did you know that a knight is better placed in the center of the board?"
            ),
            _(
                "Did you know that moving the queen at the very beginning of a game does not offer any particular advantage?"
            ),
            _(
                "Did you know that having two-colored bishops working together is very powerful?"
            ),
            _("Did you know that the rooks are generally engaged late in the game?"),
            _(
                "Did you know that the king can move across two squares under certain conditions? This is called "
                "castling"
                "."
            ),
            _(
                "Did you know that the number of possible chess games exceeds the number of atoms in the Universe?"
            ),
            # General UI
            _(
                "You can start a new game with <b>Game</b> > <b>New Game</b>, then choose the <b>Players</b>, <b>Time Control</b> and <b>Chess Variants</b>."
            ),
            _(
                "You can choose from 20 different difficulties to play against the computer. It will mainly affect the engines available time to think."
            ),
            _(
                "The level 20 engine difficulty gives full autonomy to the chess engine in managing its own time during the game."
            ),
            _(
                "To save a game use <b>Game</b> > <b>Save Game As</b>, provide the filename, and choose where you want it to be saved. At the bottom choose the extension type of the file, and <b>Save</b>."
            ),
            _(
                " "
                "Calling the flag"
                " is the termination of the current game when the time of your opponent is over. If the clock is on your side, click on the menu item <b>Actions</b> > <b>Call Flag</b> to claim the victory."
            ),
            _(
                "Press <b>Ctrl+Z</b> to ask your opponent to rollback the last played move. Against a computer or for an unrated game, undoing is generally automatically accepted."
            ),
            _(
                "To play in <b>Fullscreen mode</b>, just press the key <b>F11</b>. Press it again to exit this mode."
            ),
            _(
                "Many sounds are emitted by PyChess while you are playing if you activate them in the preferences: <b>Settings</b> > <b>Preferences</b> > <b>Sound tab</b> > <b>Use sounds in PyChess</b>."
            ),
            _(
                "Did you know that a game is generally finished after 20 to 40 moves per player? The estimated duration of a game is displayed when you configure a new game."
            ),
            _(
                "The standard file format to manage chess games is <b>PGN</b>. It stands for "
                "Portable Game Notation"
                ". Do not get confused with PNG which is a common file format to store drawings and pictures."
            ),
            _(
                "You can share a position by using the exchange format <b>FEN</b>, which stands for "
                "Forsyth-Edwards Notation"
                ". This format is also adapted for the chess variants."
            ),
            # Analysis
            _(
                "You must define a chess engine in the preferences in order to use the local chess analysis. By default, PyChess recommends the use of the free engine named Stockfish which is renowned to be among the strongest engines in the world."
            ),
            _(
                "<b>Hint mode</b> analyzes your game to show you the best current move. Enable it with the shortcut <b>Ctrl+H</b> or from the <b>View</b> menu."
            ),
            _(
                "<b>Spy mode</b> analyzes the threats and shows the best move that your opponent would play if it was his turn. Enable it with the shortcut <b>Ctrl+Y</b> or from the <b>View</b> menu."
            ),
            _(
                "<b>Ponder</b> is an option available in some chess engines that allows thinking when it is not the engine's turn. Using this feature will consume more resources on your computer."
            ),
            _(
                "<b>MultiPV</b> is an option of some chess engines that shows other possible good moves. They are displayed in the panel <b>Hints</b>. The value can be adapted from that panel with a double-click on the displayed figure."
            ),
            _(
                "You cannot use the local chess analysis mode while you are playing an unterminated game over the Internet. Else you would be a cheater."
            ),
            _(
                "An evaluation of +2.3 is an advantage for White of more than 2 pawns, even if White and Black have the same number of pawns. The position of all the pieces and their mobility are some of the factors that contribute to the engine score."
            ),
            _(
                "PyChess includes a chess engine that offers an evaluation for any chess position. Winning against PyChess engine is a coherent way to succeed in chess and improve your skills."
            ),
            _(
                "The rating is your strength: 1500 is a good average player, 2000 is a national champion and 2800 is the best human chess champion. You can view the ratings of the players via the <b>Properties</b> in the <b>Game</b> menu. The difference of points gives you your chance to win and the projected evolution of your rating. If your rating is provisional, it is appended with a question mark '?', for example "
                "1399?"
                "."
            ),
            _(
                "Several rating systems exist to evaluate your skills in chess. The most common one is called ELO (from its creator Arpad Elo) which was established in 1970. Schematically, the concept is to engage +/- 20 points for a game and that you will win or lose proportionally to the difference of ELO points you have with your opponent."
            ),
            _(
                "Each chess engine has its own evaluation function. It is normal to get varying scores for the same position when using different engines."
            ),
            # Opening book and EGDB
            _(
                "The opening book gives you the first few moves that are considered to be good from a theoretical perspective. You are free to play any other legal move."
            ),
            _(
                "The <b>Gaviota tables</b> are precalculated positions that tell the final outcome of the current game in terms of win, loss or draw."
            ),
            _(
                "Did you know that your computer is too small to store a 7-piece endgame database? That's why the Gaviota tablebase is limited to 5 pieces."
            ),
            _(
                "A tablebase can be connected either to PyChess, or to a compatible chess engine."
            ),
            _(
                "The <b>DTZ</b> is the "
                "distance to zero"
                ", so the remaining possible moves to end the game in a tie as soon as possible."
            ),
            # Variant chess
            _(
                "The chess variants consist of changing the start position, the rules of the game, the types of the pieces... The gameplay is totally modified, so you must use specialized chess engines to play against the computer."
            ),
            _(
                "In Chess960, the order of the main pieces are shuffled in a precise manner. Therefore, you cannot use the opening book and you should change your tactical habits."
            ),
            _(
                "When playing crazyhouse chess, the captured pieces change of ownership and can reappear on the board at a later turn."
            ),
            _(
                "Suicide chess, giveaway chess or antichess are all the same variant: you must give your pieces to your opponent by forcing the captures like at draughts (aka checkers). The outcome of the game can change completely if you make an incorrect move."
            ),
            _(
                "Playing horde in PyChess consists in destroying an onslaught of 36 white pawns with a normal set of black pieces."
            ),
            _(
                "You might be interested in playing "
                "King of the hill"
                " where the goal is to either checkmate or to place your king in the middle of the board, instead of protecting it in a corner of the board as usual."
            ),
            _(
                "A lot of fun is offered by atomic chess where one destroys all the surrounding main pieces at each capture."
            ),
            _(
                "The experienced chess players can use blind pieces by starting a new variant game."
            ),
            # Internet chess
            _(
                "You should sign up online to play on an Internet chess server, so that you can find your games later and see the evolution of your rating. In <b>Preferences</b> PyChess provides the possibility to save your played games locally as well."
            ),
            _(
                "Time compensation is a feature that doesn't waste your clock time because of the latency of your Internet connection. The module can be downloaded from the menu <b>Edit</b> > <b>Externals</b>."
            ),
            _(
                "You can play against chess engines on an Internet chess server. Use the filter to include or exclude them from the available players."
            ),
            _(
                "The communication with an Internet chess server is not standardized. Therefore, you can only connect to the supported chess servers in PyChess, like freechess.org or chessclub.com"
            ),
            _(
                "PyChess is able to download chess games remotely from the following websites:"
            )
            + "\n\n"
            + ", ".join(get_internet_game_providers()),
            # Externals
            _(
                "PyChess uses the external module Scoutfish to evaluate the chess databases. For example, it is possible to extract the games where some pieces are in precise count or positions."
            ),
            _(
                "Parser/ChessDB is an external module used by PyChess to show the expected outcome for a given position."
            ),
            _(
                "SQLite is an internal module used to describe the loaded PGN files, so that PyChess can retrieve the games very quickly during a search."
            ),
            _(
                "PyChess generates three information files when a PGN file is opened: .sqlite (description), .scout (positions), .bin (book and outcomes). These files can be removed manually if desired."
            ),
            # Lessons
            _(
                "PyChess provides offline lessons to learn chess so that you will never be disappointed if you have no Internet connection."
            ),
            _(
                "To start Learning, click on the <b>Book icon</b> available on the <b>Welcome</b> screen. Or choose a category next to the button to start an activity directly."
            ),
            _(
                "The <b>Lectures</b> are commented games to learn step-by-step strategy and principles of some chess techniques. Just watch, read and learn."
            ),
            _(
                "Whatever the number of pawns, an <b>end-game</b> starts when the board consists of certain main pieces : 1 rook vs 1 bishop, 1 queen versus 2 rooks, etc... Knowing the right endgame moves will help you not miss a checkmate!"
            ),
            _(
                "A <b>Puzzle</b> is a set of simple positions classified by themes where you guess the best moves. It helps you understand patterns to perform an accurate attack or defense."
            ),
            _(
                "A <b>Lesson</b> is a complex study that explains the tactics for a given position. It is common to view circles and arrows over the board to focus on the behavior of the pieces, the threats, etc..."
            ),
        ]
        self.tips_seed = conf.get("tips_seed")
        if self.tips_seed == 0:  # Forbidden value
            self.tips_seed = 123456789 + randrange(876543210)
            conf.set("tips_seed", self.tips_seed)
        self.tips_curindex = conf.get("tips_index")
        self.shuffleTips()

    def xorshift(self):
        self.tips_seed ^= (
            (self.tips_seed << 13) ^ (self.tips_seed >> 17) ^ (self.tips_seed << 5)
        )
        return self.tips_fixed + self.tips_seed % (len(self.tips) - self.tips_fixed)

    def shuffleTips(self):
        # Because of the fixed seed, the same shuffled list will be returned.
        # The idea is to manage all the tips above by category but to display
        # them once from a random order that can be reproduced at every run.
        for i in range(len(self.tips) - self.tips_fixed):
            p1 = self.xorshift()
            p2 = self.xorshift()
            if p1 != p2:
                tmp = self.tips[p1]
                self.tips[p1] = self.tips[p2]
                self.tips[p2] = tmp

    def show(self):
        self.set_currentIndex(self.tips_curindex)
        self.widgets["window1"].show()
        self.widgets["window1"].present()

    def set_currentIndex(self, value):
        if len(self.tips) == 0:
            return
        if value < 0:
            value = len(self.tips) - 1
        elif value >= len(self.tips):
            value = 0
        self.tips_curindex = value
        conf.set(
            "tips_index", self.tips_curindex + 1
        )  # To get the next message loaded at the next run
        self.widgets["tipfield"].set_markup(self.tips[value])



"""This module handles the tabbed layout in PyChess"""

import sys
from collections import defaultdict

from gi.repository import Gtk, GObject

import pychess
from .BoardControl import BoardControl
from .ChessClock import ChessClock
from .MenuItemsDict import MenuItemsDict
from pychess.System import conf

from pychess.System.Log import log
from pychess.Utils.IconLoader import get_pixbuf
from pychess.Utils.const import (
    REMOTE,
    UNFINISHED_STATES,
    PAUSED,
    RUNNING,
    LOCAL,
    WHITE,
    BLACK,
    ACTION_MENU_ITEMS,
    DRAW,
    UNDOABLE_STATES,
    HINT,
    SPY,
    WHITEWON,
    MENU_ITEMS,
    BLACKWON,
    DROP,
    FAN_PIECES,
    TOOL_CHESSDB,
    TOOL_SCOUTFISH,
)
from pychess.Utils.GameModel import GameModel
from pychess.Utils.Move import listToMoves
from pychess.Utils.lutils import lmove
from pychess.Utils.lutils.lmove import ParsingError
from pychess.Utils.logic import playerHasMatingMaterial, isClaimableDraw
from pychess.ic import get_infobarmessage_content, get_infobarmessage_content2
from pychess.ic.FICSObjects import get_player_tooltip_text
from pychess.ic.ICGameModel import ICGameModel
from pychess.widgets import createImage, createAlignment, gtk_close
from pychess.widgets.InfoBar import (
    InfoBarNotebook,
    InfoBarMessage,
    InfoBarMessageButton,
)
from pychess.perspectives import perspective_manager


light_on = get_pixbuf("glade/16x16/weather-clear.png")
light_off = get_pixbuf("glade/16x16/weather-clear-night.png")

widgets = None


def setWidgets(w):
    global widgets
    widgets = w
    pychess.widgets.main_window = widgets["main_window"]


def getWidgets():
    return widgets


class GameWidget(GObject.GObject):
    __gsignals__ = {
        "game_close_clicked": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "title_changed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "closed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, gamemodel, perspective):
        GObject.GObject.__init__(self)
        self.gamemodel = gamemodel
        self.perspective = perspective
        self.cids = {}
        self.closed = False

        # InfoBarMessage with rematch, undo or observe buttons
        self.game_ended_message = None

        (
            self.tabcontent,
            white_label,
            black_label,
            self.game_info_label,
        ) = self.initTabcontents()
        self.boardvbox, self.board, self.infobar, self.clock = self.initBoardAndClock(
            self.gamemodel
        )
        self.stat_hbox = self.initButtons(self.board)

        self.player_name_labels = (white_label, black_label)
        self.infobar.connect("hide", self.infobar_hidden)

        self.notebookKey = Gtk.Alignment()
        self.menuitems = MenuItemsDict()

        self.gamemodel_cids = [
            self.gamemodel.connect_after("game_started", self.game_started),
            self.gamemodel.connect_after("game_ended", self.game_ended),
            self.gamemodel.connect_after("game_changed", self.game_changed),
            self.gamemodel.connect("game_paused", self.game_paused),
            self.gamemodel.connect("game_resumed", self.game_resumed),
            self.gamemodel.connect("moves_undone", self.moves_undone),
            self.gamemodel.connect("game_unended", self.game_unended),
            self.gamemodel.connect("game_saved", self.game_saved),
            self.gamemodel.connect("players_changed", self.players_changed),
            self.gamemodel.connect("analyzer_added", self.analyzer_added),
            self.gamemodel.connect("analyzer_removed", self.analyzer_removed),
            self.gamemodel.connect("message_received", self.message_received),
        ]
        self.players_changed(self.gamemodel)

        self.notify_cids = [
            conf.notify_add("showFICSgameno", self.on_show_fics_gameno),
        ]

        if self.gamemodel.display_text:
            if isinstance(self.gamemodel, ICGameModel) and conf.get("showFICSgameno"):
                self.game_info_label.set_text(
                    f"{self.display_text} [{self.gamemodel.ficsgame.gameno}]"
                )
            else:
                self.game_info_label.set_text(self.display_text)
        if self.gamemodel.timed:
            self.cids[self.gamemodel.timemodel] = self.gamemodel.timemodel.connect(
                "zero_reached", self.zero_reached
            )

        self.connections = defaultdict(list)
        if isinstance(self.gamemodel, ICGameModel):
            self.connections[self.gamemodel.connection.bm].append(
                self.gamemodel.connection.bm.connect(
                    "player_lagged", self.player_lagged
                )
            )
            self.connections[self.gamemodel.connection.bm].append(
                self.gamemodel.connection.bm.connect(
                    "opp_not_out_of_time", self.opp_not_out_of_time
                )
            )
        self.cids[self.board.view] = self.board.view.connect(
            "shownChanged", self.shownChanged
        )

        if isinstance(self.gamemodel, ICGameModel):
            self.gamemodel.gmwidg_ready.set()

    def _del(self):
        if self.gamemodel.offline_lecture:
            self.gamemodel.lecture_exit_event.set()

        for obj in self.cids:
            if obj.handler_is_connected(self.cids[obj]):
                log.debug("GameWidget._del: disconnecting %s" % repr(obj))
                obj.disconnect(self.cids[obj])
        self.cids = {}

        for obj in self.connections:
            for handler_id in self.connections[obj]:
                if obj.handler_is_connected(handler_id):
                    obj.disconnect(handler_id)
        self.connections = {}

        for cid in self.gamemodel_cids:
            self.gamemodel.disconnect(cid)

        for cid in self.notify_cids:
            conf.notify_remove(cid)

        self.board._del()

        if self.game_ended_message is not None:
            self.game_ended_message.callback = None

    def on_show_fics_gameno(self, *args):
        """Checks the configuration / preferences to see if the FICS
        game number should be displayed next to player names.
        """
        if isinstance(self.gamemodel, ICGameModel) and conf.get("showFICSgameno"):
            self.game_info_label.set_text(" [%s]" % self.gamemodel.ficsgame.gameno)
        else:
            self.game_info_label.set_text("")

    def infront(self):
        for menuitem in self.menuitems:
            self.menuitems[menuitem].update()

        for widget in MENU_ITEMS:
            if widget in self.menuitems:
                continue
            elif widget == "show_sidepanels" and isDesignGWShown():
                getWidgets()[widget].set_property("sensitive", False)
            else:
                getWidgets()[widget].set_property("sensitive", True)

        # Change window title
        getWidgets()["main_window"].set_title(
            self.display_text + (" - " if self.display_text != "" else "") + "PyChess"
        )

    def _update_menu_abort(self):
        if self.gamemodel.hasEnginePlayer():
            self.menuitems["abort"].sensitive = True
            self.menuitems["abort"].tooltip = ""
        elif self.gamemodel.isObservationGame():
            self.menuitems["abort"].sensitive = False
        elif (
            isinstance(self.gamemodel, ICGameModel)
            and self.gamemodel.status in UNFINISHED_STATES
        ):
            if self.gamemodel.ply < 2:
                self.menuitems["abort"].label = _("Abort")
                self.menuitems["abort"].tooltip = _(
                    "This game can be automatically aborted without rating loss because \
                      there has not yet been two moves made"
                )
            else:
                self.menuitems["abort"].label = _("Offer Abort")
                self.menuitems["abort"].tooltip = _(
                    "Your opponent must agree to abort the game because there \
                      has been two or more moves made"
                )
            self.menuitems["abort"].sensitive = True
        else:
            self.menuitems["abort"].sensitive = False
            self.menuitems["abort"].tooltip = ""

    def _update_menu_adjourn(self):
        self.menuitems["adjourn"].sensitive = (
            isinstance(self.gamemodel, ICGameModel)
            and self.gamemodel.status in UNFINISHED_STATES
            and not self.gamemodel.isObservationGame()
            and not self.gamemodel.hasGuestPlayers()
        )

        if (
            isinstance(self.gamemodel, ICGameModel)
            and self.gamemodel.status in UNFINISHED_STATES
            and not self.gamemodel.isObservationGame()
            and self.gamemodel.hasGuestPlayers()
        ):
            self.menuitems["adjourn"].tooltip = _(
                "This game can not be adjourned because one or both players are guests"
            )
        else:
            self.menuitems["adjourn"].tooltip = ""

    def _update_menu_draw(self):
        self.menuitems["draw"].sensitive = (
            self.gamemodel.status in UNFINISHED_STATES
            and not self.gamemodel.isObservationGame()
        )

        def can_win(color):
            if self.gamemodel.timed:
                return (
                    playerHasMatingMaterial(self.gamemodel.boards[-1], color)
                    and self.gamemodel.timemodel.getPlayerTime(color) > 0
                )
            else:
                return playerHasMatingMaterial(self.gamemodel.boards[-1], color)

        if isClaimableDraw(self.gamemodel.boards[-1]) or not (
            can_win(self.gamemodel.players[0].color)
            or can_win(self.gamemodel.players[1].color)
        ):
            self.menuitems["draw"].label = _("Claim Draw")

    def _update_menu_resign(self):
        self.menuitems["resign"].sensitive = (
            self.gamemodel.status in UNFINISHED_STATES
            and not self.gamemodel.isObservationGame()
        )

    def _update_menu_pause_and_resume(self):
        def game_is_pausable():
            if self.gamemodel.isEngine2EngineGame() or (
                self.gamemodel.hasLocalPlayer()
                and (
                    self.gamemodel.isLocalGame()
                    or (
                        isinstance(self.gamemodel, ICGameModel)
                        and self.gamemodel.ply > 1
                    )
                )
            ):
                if sys.platform == "win32" and self.gamemodel.hasEnginePlayer():
                    return False
                else:
                    return True
            else:
                return False

        self.menuitems["pause1"].sensitive = (
            self.gamemodel.status == RUNNING and game_is_pausable()
        )
        self.menuitems["resume1"].sensitive = (
            self.gamemodel.status == PAUSED and game_is_pausable()
        )
        # TODO: if IC game is over and game ended in adjournment
        #       and opponent is available, enable Resume

    def _update_menu_undo(self):
        if self.gamemodel.isObservationGame():
            self.menuitems["undo1"].sensitive = False
        elif isinstance(self.gamemodel, ICGameModel):
            if self.gamemodel.status in UNFINISHED_STATES and self.gamemodel.ply > 0:
                self.menuitems["undo1"].sensitive = True
            else:
                self.menuitems["undo1"].sensitive = False
        elif self.gamemodel.ply > 0 and self.gamemodel.status in UNDOABLE_STATES + (
            RUNNING,
        ):
            self.menuitems["undo1"].sensitive = True
        else:
            self.menuitems["undo1"].sensitive = False

    def _update_menu_ask_to_move(self):
        if self.gamemodel.isObservationGame():
            self.menuitems["ask_to_move"].sensitive = False
        elif isinstance(self.gamemodel, ICGameModel):
            self.menuitems["ask_to_move"].sensitive = False
        elif (
            self.gamemodel.waitingplayer.__type__ == LOCAL
            and self.gamemodel.status in UNFINISHED_STATES
            and self.gamemodel.status != PAUSED
        ):
            self.menuitems["ask_to_move"].sensitive = True
        else:
            self.menuitems["ask_to_move"].sensitive = False

    def _showHolding(self, holding):
        figurines = ["", ""]
        for color in (BLACK, WHITE):
            for piece in holding[color].keys():
                count = holding[color][piece]
                figurines[color] += (
                    " " if count == 0 else FAN_PIECES[color][piece] * count
                )
        print(figurines[BLACK] + "   " + figurines[WHITE])

    def shownChanged(self, boardview, shown):
        # Help crazyhouse testing
        #    if self.gamemodel.boards[-1].variant == CRAZYHOUSECHESS:
        #    holding = self.gamemodel.getBoardAtPly(shown, boardview.variation).board.holding
        #    self._showHolding(holding)

        if (
            self.gamemodel.timemodel.hasTimes
            and (
                self.gamemodel.endstatus
                or self.gamemodel.status in (DRAW, WHITEWON, BLACKWON)
            )
            and boardview.shownIsMainLine()
        ):
            wmovecount, color = divmod(shown + 1, 2)
            bmovecount = wmovecount - 1 if color == WHITE else wmovecount
            if self.gamemodel.timemodel.hasBWTimes(bmovecount, wmovecount):
                self.clock.update(wmovecount, bmovecount)

        self.on_shapes_changed(self.board)

    def game_started(self, gamemodel):
        if self.gamemodel.isLocalGame():
            self.menuitems["abort"].label = _("Abort")
        self._update_menu_abort()
        self._update_menu_adjourn()
        self._update_menu_draw()
        if self.gamemodel.isLocalGame():
            self.menuitems["pause1"].label = _("Pause")
            self.menuitems["resume1"].label = _("Resume")
        else:
            self.menuitems["pause1"].label = _("Offer Pause")
            self.menuitems["resume1"].label = _("Offer Resume")
        self._update_menu_pause_and_resume()
        self._update_menu_resign()
        if self.gamemodel.isLocalGame():
            self.menuitems["undo1"].label = _("Undo")
        else:
            self.menuitems["undo1"].label = _("Offer Undo")
        self._update_menu_undo()
        self._update_menu_ask_to_move()

        if isinstance(gamemodel, ICGameModel) and not gamemodel.isObservationGame():
            for item in self.menuitems:
                if item in self.menuitems.ANAL_MENU_ITEMS:
                    self.menuitems[item].sensitive = False

        if not gamemodel.timed and not gamemodel.timemodel.hasTimes:
            try:
                self.boardvbox.remove(self.clock.get_parent())
            except TypeError:
                # no clock
                pass

    def game_ended(self, gamemodel, reason):
        for item in self.menuitems:
            if item in self.menuitems.ANAL_MENU_ITEMS:
                self.menuitems[item].sensitive = True
            elif item not in self.menuitems.VIEW_MENU_ITEMS:
                self.menuitems[item].sensitive = False
        self._update_menu_undo()
        self._set_arrow(HINT, None)
        self._set_arrow(SPY, None)
        return False

    def game_changed(self, gamemodel, ply):
        """This runs when the game changes. It updates everything."""
        self._update_menu_abort()
        self._update_menu_ask_to_move()
        self._update_menu_draw()
        self._update_menu_pause_and_resume()
        self._update_menu_undo()
        if isinstance(gamemodel, ICGameModel):  # on FICS game board change update allob
            if gamemodel.connection is not None and not gamemodel.connection.ICC:
                allob = "allob " + str(gamemodel.ficsgame.gameno)
                gamemodel.connection.client.run_command(allob)

        for analyzer_type in (HINT, SPY):
            # only clear arrows if analyzer is examining the last position
            if (
                analyzer_type in gamemodel.spectators
                and gamemodel.spectators[analyzer_type].board == gamemodel.boards[-1]
            ):
                self._set_arrow(analyzer_type, None)
        self.name_changed(gamemodel.players[0])  # We may need to add * to name

        if gamemodel.isObservationGame() and not self.isInFront():
            self.light_on_off(True)

        # print(gamemodel.waitingplayer, gamemodel.waitingplayer.__type__)
        if not gamemodel.isPlayingICSGame():
            self.clearMessages()

        return False

    def game_saved(self, gamemodel, uri):
        """Run when the game is saved. Will remove * from title."""
        self.name_changed(gamemodel.players[0])  # We may need to remove * in name
        return False

    def game_paused(self, gamemodel):
        self._update_menu_pause_and_resume()
        self._update_menu_undo()
        self._update_menu_ask_to_move()
        return False

    def game_resumed(self, gamemodel):
        self._update_menu_pause_and_resume()
        self._update_menu_undo()
        self._update_menu_ask_to_move()
        return False

    def moves_undone(self, gamemodel, moves):
        self.game_changed(gamemodel, 0)
        return False

    def game_unended(self, gamemodel):
        self._update_menu_abort()
        self._update_menu_adjourn()
        self._update_menu_draw()
        self._update_menu_pause_and_resume()
        self._update_menu_resign()
        self._update_menu_undo()
        self._update_menu_ask_to_move()
        return False

    def _set_arrow(self, analyzer_type, coordinates):
        if self.gamemodel.isPlayingICSGame():
            return

        try:
            if analyzer_type == HINT:
                self.board.view._setGreenarrow(coordinates)
            else:
                self.board.view._setRedarrow(coordinates)
        except AttributeError:
            # first cord is None
            pass

    def _on_analyze(self, analyzer, analysis, analyzer_type):
        if self.board.view.animating:
            return

        if not self.menuitems[analyzer_type + "_mode"].active:
            return

        if len(analysis) >= 1 and analysis[0] is not None:
            ply, movstrs, score, depth, nps = analysis[0]
            board = analyzer.board
            try:
                moves = listToMoves(board, movstrs, validate=True)
            except ParsingError as e:
                # ParsingErrors may happen when parsing "old" lines from
                # analyzing engines, which haven't yet noticed their new tasks
                log.debug(
                    "GameWidget._on_analyze(): Ignored (%s) from analyzer: ParsingError%s"
                    % (" ".join(movstrs), e)
                )
                return

            if moves and (
                self.gamemodel.curplayer.__type__ == LOCAL
                or [player.__type__ for player in self.gamemodel.players]
                == [REMOTE, REMOTE]
                or self.gamemodel.status not in UNFINISHED_STATES
            ):
                if moves[0].flag == DROP:
                    piece = lmove.FCORD(moves[0].move)
                    color = board.color if analyzer_type == HINT else 1 - board.color
                    cord0 = board.getHoldingCord(color, piece)
                    self._set_arrow(analyzer_type, (cord0, moves[0].cord1))
                else:
                    self._set_arrow(analyzer_type, moves[0].cords)
            else:
                self._set_arrow(analyzer_type, None)
        return False

    def analyzer_added(self, gamemodel, analyzer, analyzer_type):
        self.cids[analyzer] = analyzer.connect(
            "analyze", self._on_analyze, analyzer_type
        )
        # self.menuitems[analyzer_type + "_mode"].active = True
        self.menuitems[analyzer_type + "_mode"].sensitive = True
        return False

    def analyzer_removed(self, gamemodel, analyzer, analyzer_type):
        self._set_arrow(analyzer_type, None)
        # self.menuitems[analyzer_type + "_mode"].active = False
        self.menuitems[analyzer_type + "_mode"].sensitive = False

        try:
            if analyzer.handler_is_connected(self.cids[analyzer]):
                analyzer.disconnect(self.cids[analyzer])
            del self.cids[analyzer]
        except KeyError:
            pass

        return False

    def show_arrow(self, analyzer, analyzer_type):
        self.menuitems[analyzer_type + "_mode"].active = True
        self._on_analyze(analyzer, analyzer.getAnalysis(), analyzer_type)
        return False

    def hide_arrow(self, analyzer, analyzer_type):
        self.menuitems[analyzer_type + "_mode"].active = False
        self._set_arrow(analyzer_type, None)
        return False

    def player_display_text(self, color, with_elo):
        text = ""
        if isinstance(self.gamemodel, ICGameModel):
            if self.gamemodel.ficsplayers:
                text = self.gamemodel.ficsplayers[color].name
                if (
                    self.gamemodel.connection.username
                    == self.gamemodel.ficsplayers[color].name
                ) and self.gamemodel.ficsplayers[color].isGuest():
                    text += " (Player)"
        else:
            if self.gamemodel.players:
                text = repr(self.gamemodel.players[color])
        if with_elo:
            elo = self.gamemodel.tags.get("WhiteElo" if color == WHITE else "BlackElo")
            if elo not in [None, "", "?", "0", 0]:
                text += " (%s)" % str(elo)
        return text

    @property
    def display_text(self):
        if not self.gamemodel.players:
            return ""
        """This will give you the name of the game."""
        vs = " - "
        t = vs.join(
            (
                self.player_display_text(WHITE, True),
                self.player_display_text(BLACK, True),
            )
        )
        return t

    def players_changed(self, gamemodel):
        log.debug("GameWidget.players_changed: starting %s" % repr(gamemodel))
        for player in gamemodel.players:
            self.name_changed(player)
            # Notice that this may connect the same player many times. In
            # normal use that shouldn't be a problem.
            self.cids[player] = player.connect("name_changed", self.name_changed)
        log.debug("GameWidget.players_changed: returning")

    def name_changed(self, player):
        log.debug("GameWidget.name_changed: starting %s" % repr(player))
        color = self.gamemodel.color(player)

        if self.gamemodel is None:
            return
        name = self.player_display_text(color, False)
        self.gamemodel.tags["White" if color == WHITE else "Black"] = name
        self.player_name_labels[color].set_text(name)
        if isinstance(self.gamemodel, ICGameModel) and player.__type__ == REMOTE:
            self.player_name_labels[color].set_tooltip_text(
                get_player_tooltip_text(
                    self.gamemodel.ficsplayers[color], show_status=False
                )
            )

        self.emit("title_changed", self.display_text)
        log.debug("GameWidget.name_changed: returning")

    def message_received(self, gamemodel, name, msg):
        if gamemodel.isObservationGame() and not self.isInFront():
            text = self.game_info_label.get_text()
            self.game_info_label.set_markup(
                '<span color="red" weight="bold">%s</span>' % text
            )

    def zero_reached(self, timemodel, color):
        if self.gamemodel.status not in UNFINISHED_STATES:
            return

        if (
            self.gamemodel.players[0].__type__ == LOCAL
            and self.gamemodel.players[1].__type__ == LOCAL
        ):
            self.menuitems["call_flag"].sensitive = True
            return

        for player in self.gamemodel.players:
            opplayercolor = BLACK if player == self.gamemodel.players[WHITE] else WHITE
            if player.__type__ == LOCAL and opplayercolor == color:
                log.debug(
                    "gamewidget.zero_reached: LOCAL player=%s, color=%s"
                    % (repr(player), str(color))
                )
                self.menuitems["call_flag"].sensitive = True
                break

    def player_lagged(self, bm, player):
        if player in self.gamemodel.ficsplayers:
            content = get_infobarmessage_content(
                player,
                _(" has lagged for 30 seconds"),
                self.gamemodel.ficsgame.game_type,
            )

            def response_cb(infobar, response, message):
                message.dismiss()
                return False

            message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
            message.add_button(
                InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
            )
            self.showMessage(message)
        return False

    def opp_not_out_of_time(self, bm):
        if self.gamemodel is not None and self.gamemodel.remote_player.time <= 0:
            content = get_infobarmessage_content2(
                self.gamemodel.remote_ficsplayer,
                _(" is lagging heavily but hasn't disconnected"),
                _("Continue to wait for opponent, or try to adjourn the game?"),
                gametype=self.gamemodel.ficsgame.game_type,
            )

            def response_cb(infobar, response, message):
                if response == 2:
                    self.gamemodel.connection.client.run_command("adjourn")
                message.dismiss()
                return False

            message = InfoBarMessage(Gtk.MessageType.QUESTION, content, response_cb)
            message.add_button(InfoBarMessageButton(_("Wait"), Gtk.ResponseType.CANCEL))
            message.add_button(InfoBarMessageButton(_("Adjourn"), 2))
            self.showMessage(message)
        return False

    def on_game_close_clicked(self, button):
        log.debug("gamewidget.on_game_close_clicked %s" % button)
        self.emit("game_close_clicked")

    def initTabcontents(self):
        tabcontent = createAlignment(0, 0, 0, 0)
        hbox = Gtk.HBox()
        hbox.set_spacing(4)
        hbox.pack_start(createImage(light_off), False, True, 0)
        close_button = Gtk.Button()
        close_button.set_property("can-focus", False)
        close_button.add(createImage(gtk_close))
        close_button.set_relief(Gtk.ReliefStyle.NONE)
        close_button.set_size_request(20, 18)

        self.cids[close_button] = close_button.connect(
            "clicked", self.on_game_close_clicked
        )

        hbox.pack_end(close_button, False, True, 0)
        text_hbox = Gtk.HBox()
        white_label = Gtk.Label(label="")
        text_hbox.pack_start(white_label, False, True, 0)
        text_hbox.pack_start(Gtk.Label(label=" - "), False, True, 0)
        black_label = Gtk.Label(label="")
        text_hbox.pack_start(black_label, False, True, 0)
        gameinfo_label = Gtk.Label(label="")
        text_hbox.pack_start(gameinfo_label, False, True, 0)
        #        label.set_alignment(0,.7)
        hbox.pack_end(text_hbox, True, True, 0)
        tabcontent.add(hbox)
        tabcontent.show_all()  # Gtk doesn't show tab labels when the rest is
        return tabcontent, white_label, black_label, gameinfo_label

    def initBoardAndClock(self, gamemodel):
        boardvbox = Gtk.VBox()
        boardvbox.set_spacing(2)
        infobar = InfoBarNotebook("gamewidget_infobar")

        ccalign = createAlignment(0, 0, 0, 0)
        cclock = ChessClock()
        cclock.setModel(gamemodel.timemodel)
        ccalign.add(cclock)
        ccalign.set_size_request(-1, 32)
        boardvbox.pack_start(ccalign, False, True, 0)

        actionMenuDic = {}
        for item in ACTION_MENU_ITEMS:
            actionMenuDic[item] = widgets[item]

        if self.gamemodel.offline_lecture:
            preview = True
        else:
            preview = False

        board = BoardControl(gamemodel, actionMenuDic, game_preview=preview)
        boardvbox.pack_start(board, True, True, 0)
        return boardvbox, board, infobar, cclock

    def initButtons(self, board):
        align = createAlignment(4, 0, 4, 0)
        toolbar = Gtk.Toolbar()

        image1 = get_pixbuf("glade/first.png")
        firstButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image1))
        firstButton.set_tooltip_text(_("Jump to initial position"))
        toolbar.insert(firstButton, -1)

        image2 = get_pixbuf("glade/prev.png")
        prevButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image2))
        prevButton.set_tooltip_text(_("Step back one move"))
        toolbar.insert(prevButton, -1)

        image3 = get_pixbuf("glade/mainline.png")
        mainButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image3))
        mainButton.set_tooltip_text(_("Go back to the main line"))
        toolbar.insert(mainButton, -1)

        image4 = get_pixbuf("glade/parent.png")
        upButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image4))
        upButton.set_tooltip_text(_("Go back to the parent line"))
        toolbar.insert(upButton, -1)

        image5 = get_pixbuf("glade/next.png")
        nextButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image5))
        nextButton.set_tooltip_text(_("Step forward one move"))
        toolbar.insert(nextButton, -1)

        image6 = get_pixbuf("glade/last.png")
        lastButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image6))
        lastButton.set_tooltip_text(_("Jump to latest position"))
        toolbar.insert(lastButton, -1)

        image7 = get_pixbuf("glade/find.png")
        filterButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image7))
        filterButton.set_tooltip_text(_("Find position in current database"))
        toolbar.insert(filterButton, -1)

        image8 = get_pixbuf("glade/save.png")
        self.saveButton = Gtk.ToolButton.new(Gtk.Image.new_from_pixbuf(image8))
        self.saveButton.set_tooltip_text(_("Save arrows/circles"))
        toolbar.insert(self.saveButton, -1)

        def on_clicked(button, func):
            # Prevent moving in game while lesson not finished
            if self.gamemodel.lesson_game and not self.gamemodel.solved:
                return
            else:
                func()

        self.cids[firstButton] = firstButton.connect(
            "clicked", on_clicked, self.board.view.showFirst
        )
        self.cids[prevButton] = prevButton.connect(
            "clicked", on_clicked, self.board.view.showPrev
        )
        self.cids[mainButton] = mainButton.connect(
            "clicked", on_clicked, self.board.view.backToMainLine
        )
        self.cids[upButton] = upButton.connect(
            "clicked", on_clicked, self.board.view.backToParentLine
        )
        self.cids[nextButton] = nextButton.connect(
            "clicked", on_clicked, self.board.view.showNext
        )
        self.cids[lastButton] = lastButton.connect(
            "clicked", on_clicked, self.board.view.showLast
        )
        self.cids[filterButton] = filterButton.connect(
            "clicked", on_clicked, self.find_in_database
        )
        self.cids[self.saveButton] = self.saveButton.connect(
            "clicked", on_clicked, self.save_shapes_to_pgn
        )

        self.on_shapes_changed(self.board)
        self.board.connect("shapes_changed", self.on_shapes_changed)

        tool_box = Gtk.Box()
        tool_box.pack_start(toolbar, True, True, 0)

        align.add(tool_box)
        return align

    def on_shapes_changed(self, boardcontrol):
        self.saveButton.set_sensitive(boardcontrol.view.has_unsaved_shapes)

    def find_in_database(self):
        persp = perspective_manager.get_perspective("database")
        if persp.chessfile is None:
            dialogue = Gtk.MessageDialog(
                pychess.widgets.mainwindow(),
                type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.OK,
                message_format=_("No database is currently opened."),
            )
            dialogue.run()
            dialogue.destroy()
            return

        view = self.board.view
        shown_board = self.gamemodel.getBoardAtPly(view.shown, view.shown_variation_idx)
        fen = shown_board.asFen()

        tool, found = persp.chessfile.has_position(fen)
        if not found:
            dialogue = Gtk.MessageDialog(
                pychess.widgets.mainwindow(),
                type=Gtk.MessageType.WARNING,
                buttons=Gtk.ButtonsType.OK,
                message_format=_("The position does not exist in the database."),
            )
            dialogue.run()
            dialogue.destroy()
        else:
            if tool == TOOL_CHESSDB:
                persp.chessfile.set_fen_filter(fen)
            elif tool == TOOL_SCOUTFISH:
                dialogue = Gtk.MessageDialog(
                    pychess.widgets.mainwindow(),
                    type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                    message_format=_(
                        "An approximate position has been found. Do you want to display it ?"
                    ),
                )
                response = dialogue.run()
                dialogue.destroy()
                if response != Gtk.ResponseType.YES:
                    return

                persp.chessfile.set_scout_filter({"sub-fen": fen})
            else:
                raise RuntimeError("Internal error")
            persp.gamelist.ply = view.shown
            persp.gamelist.load_games()
            perspective_manager.activate_perspective("database")

    def save_shapes_to_pgn(self):
        view = self.board.view
        shown_board = self.gamemodel.getBoardAtPly(view.shown, view.shown_variation_idx)

        for child in shown_board.board.children:
            if isinstance(child, str):
                if child.lstrip().startswith("[%csl "):
                    shown_board.board.children.remove(child)
                    self.gamemodel.needsSave = True
                elif child.lstrip().startswith("[%cal "):
                    shown_board.board.children.remove(child)
                    self.gamemodel.needsSave = True

        if view.circles:
            csl = []
            for circle in view.circles:
                csl.append(f"{circle.color}{repr(circle)}")
            shown_board.board.children = [
                "[%%csl %s]" % ",".join(csl)
            ] + shown_board.board.children
            self.gamemodel.needsSave = True

        if view.arrows:
            cal = []
            for arrow in view.arrows:
                cal.append(f"{arrow[0].color}{repr(arrow[0])}{repr(arrow[1])}")
            shown_board.board.children = [
                "[%%cal %s]" % ",".join(cal)
            ] + shown_board.board.children
            self.gamemodel.needsSave = True

        view.saved_arrows = set()
        view.saved_arrows |= view.arrows

        view.saved_circles = set()
        view.saved_circles |= view.circles

        self.on_shapes_changed(self.board)

    def light_on_off(self, on):
        child = self.tabcontent.get_child()
        if child:
            child.remove(child.get_children()[0])
            if on:
                # child.pack_start(createImage(light_on, True, True, 0), expand=False)
                child.pack_start(createImage(light_on), True, True, 0)
            else:
                # child.pack_start(createImage(light_off, True, True, 0), expand=False)
                child.pack_start(createImage(light_off), True, True, 0)
        self.tabcontent.show_all()

    def setLocked(self, locked):
        """Makes the board insensitive and turns off the tab ready indicator"""
        log.debug(
            "GameWidget.setLocked: {} locked={}".format(
                self.gamemodel.players, str(locked)
            )
        )
        self.board.setLocked(locked)
        if not self.tabcontent.get_children():
            return
        if len(self.tabcontent.get_child().get_children()) < 2:
            log.warning("GameWidget.setLocked: Not removing last tabcontent child")
            return

        self.light_on_off(not locked)

        log.debug("GameWidget.setLocked: %s: returning" % self.gamemodel.players)

    def bringToFront(self):
        self.perspective.getheadbook().set_current_page(self.getPageNumber())

    def isInFront(self):
        if not self.perspective.getheadbook():
            return False
        return self.perspective.getheadbook().get_current_page() == self.getPageNumber()

    def getPageNumber(self):
        return self.perspective.getheadbook().page_num(self.notebookKey)

    def infobar_hidden(self, infobar):
        if self == self.perspective.cur_gmwidg():
            self.perspective.notebooks["messageArea"].hide()

    def showMessage(self, message):
        self.infobar.push_message(message)
        if self == self.perspective.cur_gmwidg():
            self.perspective.notebooks["messageArea"].show()

    def replaceMessages(self, message):
        """Replace all messages with message"""
        if not self.closed:
            self.infobar.clear_messages()
            self.showMessage(message)

    def clearMessages(self):
        self.infobar.clear_messages()
        if self == self.perspective.cur_gmwidg():
            self.perspective.notebooks["messageArea"].hide()


# ###############################################################################
# Handling of the special sidepanels-design-gamewidget used in preferences     #
# ###############################################################################

designGW = None


def showDesignGW():
    global designGW
    perspective = perspective_manager.get_perspective("games")
    designGW = GameWidget(GameModel(), perspective)
    if isDesignGWShown():
        return
    getWidgets()["show_sidepanels"].set_active(True)
    getWidgets()["show_sidepanels"].set_sensitive(False)
    perspective.attachGameWidget(designGW)


def hideDesignGW():
    if isDesignGWShown():
        perspective = perspective_manager.get_perspective("games")
        perspective.delGameWidget(designGW)
    getWidgets()["show_sidepanels"].set_sensitive(True)


def isDesignGWShown():
    perspective = perspective_manager.get_perspective("games")
    return designGW in perspective.key2gmwidg.values()



from gi.repository import GObject

from pychess.Utils.const import (
    OFFERS,
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    TAKEBACK_OFFER,
)

from pychess.Utils.Offer import Offer
from pychess.System.Log import log
from pychess.ic.managers.OfferManager import OfferManager, offerTypeToStr
from pychess.ic.icc import DG_OFFERS_IN_MY_GAME, DG_MATCH, DG_MATCH_REMOVED


class ICCOfferManager(OfferManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(
            DG_OFFERS_IN_MY_GAME, self.on_icc_offers_in_my_game
        )
        self.connection.expect_dg_line(DG_MATCH, self.on_icc_match)
        self.connection.expect_dg_line(DG_MATCH_REMOVED, self.on_icc_match_removed)

        self.connection.client.run_command("set-2 %s 1" % DG_OFFERS_IN_MY_GAME)
        self.connection.client.run_command("set-2 %s 1" % DG_MATCH)
        self.connection.client.run_command("set-2 %s 1" % DG_MATCH_REMOVED)

        self.lastPly = 0
        self.offers = {}

    def on_icc_offers_in_my_game(self, data):
        log.debug("DG_OFFERS_IN_MY_GAME %s" % data)
        # gamenumber wdraw bdraw wadjourn badjourn wabort babort wtakeback btakeback
        (
            gamenumber,
            wdraw,
            bdraw,
            wadjourn,
            badjourn,
            wabort,
            babort,
            wtakeback,
            btakeback,
        ) = map(int, data.split())

        if wdraw or bdraw:
            offertype = DRAW_OFFER
        elif wadjourn or badjourn:
            offertype = ADJOURN_OFFER
        elif wabort or babort:
            offertype = ABORT_OFFER
        elif wtakeback or btakeback:
            offertype = TAKEBACK_OFFER
        else:
            log.debug(
                "ICCOfferManager.on_icc_offers_in_my_game: unknown offer data: %s"
                % data
            )
            return

        index = gamenumber * 100000 + OFFERS.index(offertype)
        if offertype == TAKEBACK_OFFER:
            parameters = wtakeback if wtakeback else btakeback
            offer = Offer(offertype, param=parameters, index=index)
        else:
            offer = Offer(offertype, index=index)
        self.offers[offer.index] = offer

        log.debug(
            "ICCOfferManager.on_icc_offers_in_my_game: emitting onOfferAdd: %s" % offer
        )
        self.emit("onOfferAdd", offer)

    def on_icc_match(self, data):
        # challenger-name challenger-rating challenger-titles
        # receiver-name   receiver-rating   receiver-titles
        # wild-number rating-type is-it-rated is-it-adjourned
        # challenger-time-control receiver-time-control
        # challenger-color-request [assess-loss assess-draw assess-win]
        # fancy-time-control
        log.debug("DG_MATCH %s" % data)

    def on_icc_match_removed(self, data):
        # challenger-name receiver-name ^Y{Explanation string^Y}
        log.debug("DG_MATCH_REMOVED %s" % data)

    def accept(self, offer):
        log.debug("OfferManager.accept: %s" % offer)
        self.connection.client.run_command("%s" % offerTypeToStr[offer.type])

    def decline(self, offer):
        log.debug("OfferManager.decline: %s" % offer)
        self.connection.client.run_command("decline %s" % offerTypeToStr[offer.type])



from gi.repository import Gtk, GLib, GdkPixbuf

from pychess.ic.FICSObjects import FICSPlayer, get_player_tooltip_text
from pychess.ic import TYPE_BLITZ, TYPE_LIGHTNING, TYPE_STANDARD, IC_STATUS_PLAYING
from pychess.System.Log import log
from pychess.System import conf, uistuff
from pychess.System.prefix import addDataPrefix
from pychess.perspectives.fics.ParrentListSection import (
    ParrentListSection,
    SEPARATOR,
    FOLLOW,
    FINGER,
    ARCHIVED,
    OBSERVE,
    CHAT,
    CHALLENGE,
)


__title__ = _("Player List")

__icon__ = addDataPrefix("glade/panel_players.svg")

__desc__ = _("List of players")


class Sidepanel(ParrentListSection):
    def load(self, widgets, connection, lounge):
        self.widgets = widgets
        self.connection = connection
        lounge.players_tab = self
        self.lounge = lounge

        __widget__ = lounge.players_list

        self.players = {}
        self.columns = {TYPE_BLITZ: 3, TYPE_STANDARD: 4, TYPE_LIGHTNING: 5}

        self.tv = widgets["playertreeview"]
        self.store = Gtk.ListStore(
            FICSPlayer, GdkPixbuf.Pixbuf, str, int, int, int, str, str
        )
        self.player_filter = self.store.filter_new()
        self.player_filter.set_visible_func(self.player_filter_func)

        self.filter_toggles = {}
        self.filter_buttons = (
            "registered_toggle",
            "guest_toggle",
            "computer_toggle",
            "titled_toggle",
        )
        for widget in self.filter_buttons:
            uistuff.keep(self.widgets[widget], widget)
            self.widgets[widget].connect("toggled", self.on_filter_button_toggled)
            initial = conf.get(widget)
            self.filter_toggles[widget] = initial
            self.widgets[widget].set_active(initial)

        self.model = self.player_filter.sort_new_with_model()
        self.tv.set_model(self.model)

        self.addColumns(
            self.tv,
            "FICSPlayer",
            "",
            _("Name"),
            _("Blitz"),
            _("Standard"),
            _("Lightning"),
            _("Status"),
            "tooltip",
            hide=[0, 7],
            pix=[1],
        )
        self.tv.set_tooltip_column(
            7,
        )
        self.tv.get_model().set_sort_func(0, self.pixCompareFunction, 1)
        try:
            self.tv.set_search_position_func(self.lowLeftSearchPosFunc, None)
        except AttributeError:
            # Unknow signal name is raised by gtk < 2.10
            pass

        connection.players.connect("FICSPlayerEntered", self.onPlayerAdded)
        connection.players.connect("FICSPlayerExited", self.onPlayerRemoved)

        widgets["private_chat_button"].connect("clicked", self.on_chat)
        widgets["private_chat_button"].set_sensitive(False)
        widgets["observe_button"].connect("clicked", self.on_observe)
        widgets["observe_button"].set_sensitive(False)

        self.tv.get_selection().connect_after("changed", self.onSelectionChanged)
        self.onSelectionChanged(None)

        self.tv.connect("button-press-event", self.button_press_event)
        self.createLocalMenu(
            (CHALLENGE, CHAT, OBSERVE, FOLLOW, SEPARATOR, FINGER, ARCHIVED)
        )

        return __widget__

    def player_filter_func(self, model, iter, data):
        player = model[iter][0]
        is_titled = player.isTitled()
        is_computer = player.isComputer()
        is_registered = (not is_titled) and (not is_computer) and (not player.isGuest())
        is_guest = (not is_titled) and (not is_computer) and (player.isGuest())
        return (
            (self.filter_toggles["computer_toggle"] and is_computer)
            or (self.filter_toggles["registered_toggle"] and is_registered)
            or (self.filter_toggles["guest_toggle"] and is_guest)
            or (self.filter_toggles["titled_toggle"] and is_titled)
        )

    def on_filter_button_toggled(self, widget):
        for button in self.filter_buttons:
            self.filter_toggles[button] = self.widgets[button].get_active()
        self.player_filter.refilter()

    def onPlayerAdded(self, players, new_players):
        # Let the hard work to be done in the helper connection thread
        np = {}
        for player in new_players:
            np[player] = (
                player,
                player.getIcon(),
                player.name + player.display_titles(),
                player.blitz,
                player.standard,
                player.lightning,
                player.display_status,
                get_player_tooltip_text(player),
            )

        def do_onPlayerAdded(players, new_players, np):
            for player in new_players:
                # log.debug("%s" % player,
                # extra={"task": (self.connection.username,
                # "PTS.onPlayerAdded")})
                if player in self.players:
                    # log.warning("%s already in self" % player,
                    # extra={"task": (self.connection.username,
                    # "PTS.onPlayerAdded")})
                    continue

                # player can leave before we finish processing "who IbslwBzSLx"
                if player not in np:
                    continue

                self.players[player] = {}

                self.players[player]["ti"] = self.store.append(np[player])
                self.players[player]["status"] = player.connect(
                    "notify::status", self.status_changed
                )
                self.players[player]["game"] = player.connect(
                    "notify::game", self.status_changed
                )
                self.players[player]["titles"] = player.connect(
                    "notify::titles", self.titles_changed
                )
                if player.game:
                    self.players[player]["private"] = player.game.connect(
                        "notify::private", self.private_changed, player
                    )
                self.players[player]["ratings"] = player.connect(
                    "ratings_changed", self.elo_changed, player
                )

            count = len(self.players)
            self.widgets["playersOnlineLabel"].set_text(_("Players: %d") % count)

            return False

        GLib.idle_add(
            do_onPlayerAdded, players, new_players, np, priority=GLib.PRIORITY_LOW
        )

    def onPlayerRemoved(self, players, player):
        def do_onPlayerRemoved(players, player):
            log.debug(
                "%s" % player,
                extra={"task": (self.connection.username, "PTS.onPlayerRemoved")},
            )
            if player not in self.players:
                return
            if self.store.iter_is_valid(self.players[player]["ti"]):
                self.store.remove(self.players[player]["ti"])
            for key in ("status", "game", "titles"):
                if player.handler_is_connected(self.players[player][key]):
                    player.disconnect(self.players[player][key])
            if (
                player.game
                and "private" in self.players[player]
                and player.game.handler_is_connected(self.players[player]["private"])
            ):
                player.game.disconnect(self.players[player]["private"])
            if player.handler_is_connected(self.players[player]["ratings"]):
                player.disconnect(self.players[player]["ratings"])
            del self.players[player]
            count = len(self.players)
            self.widgets["playersOnlineLabel"].set_text(_("Players: %d") % count)

        GLib.idle_add(do_onPlayerRemoved, players, player, priority=GLib.PRIORITY_LOW)

    def status_changed(self, player, prop):
        log.debug(
            "%s" % player,
            extra={"task": (self.connection.username, "PTS.status_changed")},
        )
        if player not in self.players:
            return

        try:
            self.store.set(self.players[player]["ti"], 6, player.display_status)
            self.store.set(
                self.players[player]["ti"], 7, get_player_tooltip_text(player)
            )
        except KeyError:
            pass

        if (
            player.status == IC_STATUS_PLAYING
            and player.game
            and "private" not in self.players[player]
        ):
            self.players[player]["private"] = player.game.connect(
                "notify::private", self.private_changed, player
            )
        elif player.status != IC_STATUS_PLAYING and "private" in self.players[player]:
            game = player.game
            if game and game.handler_is_connected(self.players[player]["private"]):
                game.disconnect(self.players[player]["private"])
            del self.players[player]["private"]

        if player == self.getSelectedPlayer():
            self.onSelectionChanged(None)

        return False

    def titles_changed(self, player, prop):
        log.debug(
            "%s" % player,
            extra={"task": (self.connection.username, "PTS.titles_changed")},
        )
        try:
            self.store.set(self.players[player]["ti"], 1, player.getIcon())
            self.store.set(
                self.players[player]["ti"], 2, player.name + player.display_titles()
            )
            self.store.set(
                self.players[player]["ti"], 7, get_player_tooltip_text(player)
            )
        except KeyError:
            pass

        return False

    def private_changed(self, game, prop, player):
        log.debug(
            "%s" % player,
            extra={"task": (self.connection.username, "PTS.private_changed")},
        )
        self.status_changed(player, prop)
        self.onSelectionChanged(self.tv.get_selection())
        return False

    def elo_changed(self, rating, prop, rating_type, player):
        log.debug(
            f"{rating} {player}",
            extra={"task": (self.connection.username, "PTS_changed")},
        )

        try:
            self.store.set(self.players[player]["ti"], 1, player.getIcon())
            self.store.set(
                self.players[player]["ti"], 7, get_player_tooltip_text(player)
            )
            self.store.set(
                self.players[player]["ti"], self.columns[rating_type], rating
            )
        except KeyError:
            pass

        return False

    def getSelectedPlayer(self):
        model, sel_iter = self.widgets["playertreeview"].get_selection().get_selected()
        if sel_iter:
            return model.get_value(sel_iter, 0)

    def onSelectionChanged(self, selection):
        player = self.getSelectedPlayer()
        user_name = self.connection.getUsername()
        self.widgets["private_chat_button"].set_sensitive(player is not None)
        self.widgets["observe_button"].set_sensitive(
            player is not None
            and player.isObservable()
            and (
                player.game is None
                or user_name not in (player.game.wplayer.name, player.game.bplayer.name)
            )
        )
        self.widgets["challengeButton"].set_sensitive(
            player is not None
            and player.isAvailableForGame()
            and player.name != user_name
        )



import sys
from math import e, floor
from random import randint

from gi.repository import Gtk, GObject
from gi.repository import Gdk

from pychess.System import uistuff, conf
from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import WHITE, DRAW, WHITEWON, BLACKWON
from pychess.Utils.lutils import leval

__title__ = _("Score")
__icon__ = addDataPrefix("glade/panel_score.svg")
__desc__ = _(
    "The score panel tries to evaluate the positions and shows you a graph of the game progress"
)


class Sidepanel:
    def load(self, gmwidg):
        self.boardview = gmwidg.board.view
        self.plot = ScorePlot(self.boardview)
        self.sw = __widget__ = Gtk.ScrolledWindow()
        __widget__.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        port = Gtk.Viewport()
        port.add(self.plot)
        port.set_shadow_type(Gtk.ShadowType.NONE)
        __widget__.add(port)
        __widget__.show_all()

        self.plot_cid = self.plot.connect("selected", self.plot_selected)
        self.cid = self.boardview.connect("shownChanged", self.shownChanged)
        self.model_cids = [
            self.boardview.model.connect_after("game_changed", self.game_changed),
            self.boardview.model.connect_after("moves_undone", self.moves_undone),
            self.boardview.model.connect_after(
                "analysis_changed", self.analysis_changed
            ),
            self.boardview.model.connect_after("game_started", self.game_started),
            self.boardview.model.connect_after(
                "game_terminated", self.on_game_terminated
            ),
        ]

        def cb_config_changed(none):
            self.fetch_chess_conf()
            self.plot.redraw()

        self.cids_conf = [conf.notify_add("scoreLinearScale", cb_config_changed)]
        self.fetch_chess_conf()

        uistuff.keepDown(__widget__)

        return __widget__

    def fetch_chess_conf(self):
        self.plot.linear_scale = conf.get("scoreLinearScale")

    def on_game_terminated(self, model):
        self.plot.disconnect(self.plot_cid)
        self.boardview.disconnect(self.cid)
        for cid in self.model_cids:
            self.boardview.model.disconnect(cid)
        for cid in self.cids_conf:
            conf.notify_remove(cid)

    def moves_undone(self, model, moves):
        for i in range(moves):
            self.plot.undo()

        # As shownChanged will normally be emitted just after game_changed -
        # if we are viewing the latest position - we can do the selection change
        # now, and thereby avoid redraw being called twice
        if self.plot.selected == model.ply - model.lowply:
            self.plot.select(model.ply - model.lowply - moves)
        self.plot.redraw()

    def game_changed(self, model, ply):
        if len(self.plot) + model.lowply > ply:
            return

        for i in range(len(self.plot) + model.lowply, ply):
            if i in model.scores:
                points = model.scores[i][1]
                points = points * -1 if i % 2 == 1 else points
            else:
                points = leval.evaluateComplete(model.getBoardAtPly(i).board, WHITE)
            self.plot.addScore(points)

        if model.status == DRAW:
            points = 0
        elif model.status == WHITEWON:
            points = sys.maxsize
        elif model.status == BLACKWON:
            points = -sys.maxsize
        else:
            if ply in model.scores:
                points = model.scores[ply][1]
                points = points * -1 if ply % 2 == 1 else points
            else:
                try:
                    points = leval.evaluateComplete(
                        model.getBoardAtPly(ply).board, WHITE
                    )
                except IndexError:
                    return
        self.plot.addScore(points)

        # As shownChanged will normally be emitted just after game_changed -
        # if we are viewing the latest position - we can do the selection change
        # now, and thereby avoid redraw being called twice
        if self.plot.selected == ply - model.lowply - 1:
            self.plot.select(ply - model.lowply)
        self.plot.redraw()

        # Uncomment this to debug eval function
        # ---
        # board = model.boards[-1].board
        # opboard = model.boards[-1].clone().board
        # opboard.setColor(1 - opboard.color)
        # material, phase = leval.evalMaterial(board)
        # if board.color == WHITE:
        #     print("material", -material)
        #     e1 = leval.evalKingTropism(board)
        #     e2 = leval.evalKingTropism(opboard)
        #     print("evaluation: %d + %d = %d " % (e1, e2, e1 + e2))
        #     p1 = leval.evalPawnStructure(board, phase)
        #     p2 = leval.evalPawnStructure(opboard, phase)
        #     print("pawns: %d + %d = %d " % (p1, p2, p1 + p2))
        #     print("knights:", -leval.evalKnights(board))
        #     print("king:", -leval.evalKing(board, phase))
        # else:
        #     print("material", material)
        #     print("evaluation:", leval.evalKingTropism(board))
        #     print("pawns:", leval.evalPawnStructure(board, phase))
        #     print("pawns2:", leval.evalPawnStructure(opboard, phase))
        #     print("pawns3:", leval.evalPawnStructure(board, phase) +
        #           leval.evalPawnStructure(opboard, phase))
        #     print("knights:", leval.evalKnights(board))
        #     print("king:", leval.evalKing(board, phase))
        # print("----------------------")

    def game_started(self, model):
        if model.lesson_game:
            return

        self.game_changed(model, model.ply)

    def shownChanged(self, boardview, shown):
        if not boardview.shownIsMainLine():
            return
        if self.plot.selected != shown:
            self.plot.select(shown - self.boardview.model.lowply)
            self.plot.redraw()

    def analysis_changed(self, gamemodel, ply):
        if self.boardview.animating:
            return

        if not self.boardview.shownIsMainLine():
            return
        if ply - gamemodel.lowply > len(self.plot.scores) - 1:
            # analysis line of yet undone position
            return

        color = (ply - 1) % 2
        score = gamemodel.scores[ply][1]
        score = score * -1 if color == WHITE else score
        self.plot.changeScore(ply - gamemodel.lowply, score)
        self.plot.redraw()

    def plot_selected(self, plot, selected):
        try:
            board = self.boardview.model.boards[selected]
        except IndexError:
            return
        self.boardview.setShownBoard(board)


class ScorePlot(Gtk.DrawingArea):
    __gtype_name__ = "ScorePlot" + str(randint(0, sys.maxsize))
    __gsignals__ = {"selected": (GObject.SignalFlags.RUN_FIRST, None, (int,))}

    def __init__(self, boardview):
        GObject.GObject.__init__(self)
        self.boardview = boardview
        self.connect("draw", self.expose)
        self.connect("button-press-event", self.press)
        self.props.can_focus = True
        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.KEY_PRESS_MASK)
        self.scores = []
        self.selected = 0

    def get_move_height(self):
        c = self.__len__()
        w = self.get_allocation().width
        if c != 0:
            w = int(floor(w / c))
        return max(min(w, 24), 1)

    def addScore(self, score):
        self.scores.append(score)

    def changeScore(self, ply, score):
        if self.scores:
            self.scores[ply] = score

    def __len__(self):
        return len(self.scores)

    def undo(self):
        del self.scores[-1]

    def select(self, index):
        self.selected = index

    def clear(self):
        del self.scores[:]

    def redraw(self):
        if self.get_window():
            a = self.get_allocation()
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = (0, 0, a.width, a.height)
            self.get_window().invalidate_rect(rect, True)
            self.get_window().process_updates(True)

    def press(self, widget, event):
        self.grab_focus()
        self.emit("selected", event.x / self.get_move_height())

    def expose(self, widget, context):
        a = widget.get_allocation()
        context.rectangle(0, 0, a.width, a.height)
        context.clip()
        self.draw(context)
        return False

    def draw(self, cr):
        m = self.boardview.model
        if m.isPlayingICSGame():
            return

        width = self.get_allocation().width
        height = self.get_allocation().height

        ########################################
        # Draw background                      #
        ########################################

        cr.set_source_rgb(1, 1, 1)
        cr.rectangle(0, 0, width, height)
        cr.fill()

        ########################################
        # Draw the actual plot (dark area)     #
        ########################################

        def sign(n):
            return n == 0 and 1 or n / abs(n)

        def mapper(score):
            if self.linear_scale:
                return min(abs(score), 800) / 800 * sign(score)  # Linear
            else:
                return (e ** (5e-4 * abs(score)) - 1) * sign(
                    score
                )  # Exponentially stretched

        if self.scores:
            cr.set_source_rgb(0, 0, 0)
            cr.move_to(0, height)
            cr.line_to(0, (height / 2.0) * (1 + mapper(self.scores[0])))
            for i, score in enumerate(self.scores):
                x = (i + 1) * self.get_move_height()
                y = (height / 2.0) * (1 + mapper(score))
                y = max(0, min(height, y))
                cr.line_to(x, y)
            cr.line_to(x, height)
            cr.fill()
        else:
            x = 0
        cr.set_source_rgb(0.9, 0.9, 0.9)
        cr.rectangle(x, 0, width, height)
        cr.fill()

        ########################################
        # Draw middle line and markers         #
        ########################################

        cr.set_line_width(0.25)
        markers = [16, -16, 8, -8, 3, -3, 0]  # centipawns
        for mark in markers:
            if mark == 0:
                cr.set_source_rgb(1, 0, 0)
            else:
                cr.set_source_rgb(0.85, 0.85, 0.85)
            y = (height / 2.0) * (1 + mapper(100 * mark))
            y = max(0, min(height, y))
            cr.move_to(0, y)
            cr.line_to(width, y)
            cr.stroke()

        ########################################
        # Draw selection                       #
        ########################################

        lw = 2
        cr.set_line_width(lw)
        s = self.get_move_height()
        x = self.selected * s
        cr.rectangle(x - lw / 2, lw / 2, s + lw, height - lw)
        found, color = self.get_style_context().lookup_color("p_bg_selected")
        cr.set_source_rgba(color.red, color.green, color.blue, 0.15)
        cr.fill_preserve()
        cr.set_source_rgb(color.red, color.green, color.blue)
        cr.stroke()



from gi.repository import GObject

from .Move import Move
from .lutils.egtb_k4it import EgtbK4kit
from .lutils.egtb_gaviota import EgtbGaviota

providers = []


class EndgameTable(GObject.GObject):
    """Wrap the low-level providers of exact endgame knowledge."""

    __gsignals__ = {
        "scored": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

        global providers
        if not providers:
            providers = [EgtbGaviota(), EgtbK4kit()]
        self.providers = providers

    def _pieceCounts(self, board):
        return sorted([bin(board.friends[i]).count("1") for i in range(2)])

    def scoreGame(self, lBoard, omitDepth=False, probeSoft=False):
        """Return result and depth to mate. (Intended for engine use.)

        lBoard: A low-level board structure
        omitDepth: Look up only the game's outcome (may save time)
        probeSoft: Fail if the probe would require disk or network access.
        Return value:
        game_result: Either WHITEWON, DRAW, BLACKWON, or (on failure) None
        depth: Depth to mate, or (if omitDepth or the game is drawn) None
        """

        piece_count = self._pieceCounts(lBoard)
        for provider in self.providers:
            if provider.supports(piece_count):
                result, depth = provider.scoreGame(lBoard, omitDepth, probeSoft)
                if result is not None:
                    return result, depth
        return None, None

    async def scoreAllMoves(self, lBoard):
        """Return each move's result and depth to mate.

        lBoard: A low-level board structure
        Return value: a list, with best moves first, of:
        move: A high-level move structure
        game_result: Either WHITEWON, DRAW, BLACKWON
        depth: Depth to mate
        """

        piece_count = self._pieceCounts(lBoard)
        for provider in self.providers:
            if provider.supports(piece_count):
                results = await provider.scoreAllMoves(lBoard)
                if results:
                    ret = []
                    for lmove, result, depth in results:
                        ret.append((Move(lmove), result, depth))
                    self.emit("scored", (lBoard, ret))
                    return ret
        return []



"""Losers Variant"""

from pychess.Utils.const import LOSERSCHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


class LosersBoard(Board):
    """:Description: The Losers variant is a game where the concept is to get rid of all your
    pieces before you opponent does. On a players turn if a piece can be taken it must be taken
    otherwise a normal chess move can be played
    """

    variant = LOSERSCHESS
    __desc__ = _(
        "FICS losers: http://www.freechess.org/Help/HelpFiles/losers_chess.html"
    )
    name = _("Losers")
    cecp_name = "losers"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD


def testKingOnly(board):
    """Checks to see if if a winning position has been acheived"""
    return bin(board.friends[board.color]).count("1") == 1



import re
from time import time

from gi.repository import GObject

from pychess.ic import (
    IC_STATUS_OFFLINE,
    IC_STATUS_ACTIVE,
    IC_STATUS_PLAYING,
    IC_STATUS_BUSY,
    GAME_TYPES_BY_FICS_NAME,
    BLKCMD_FINGER,
)
from pychess.Utils.const import WHITE, BLACK
from pychess.System.Log import log

types = "(?:blitz|standard|lightning|wild|bughouse|crazyhouse|suicide|losers|atomic)"
rated = "(rated|unrated)"
colors = r"(?:\[(white|black)\]\s?)?"
ratings = r"([\d\+\-]{1,4})"
titleslist = r"(?:GM|IM|FM|WGM|WIM|WFM|TM|SR|TD|SR|CA|C|U|D|B|T|\*)"
titles = r"((?:\(%s\))+)?" % titleslist
names = r"(\w+)%s" % titles
mf = r"(?:([mf]{1,2})\s?)?"
# FIXME: Needs support for day, hour, min, sec
times = "[, ]*".join(r"(?:(\d+) %s)?" % s for s in ("days", "hrs", "mins", "secs"))

# "73 days, 5 hrs, 55 mins"
# ('73', '5', '55', None)


class FingerObject:
    def __init__(self, name=""):
        self.__fingerTime = time()

        self.__name = name
        self.__status = None
        self.__upTime = 0
        self.__idleTime = 0
        self.__busyMessage = ""
        self.__lastSeen = 0
        self.__totalTimeOnline = 0
        self.__created = 0  # Since field from % of life online
        self.__email = ""
        self.__sanctions = ""
        self.__adminLevel = ""
        self.__timeseal = False
        self.__notes = [""] * 10
        self.__gameno = ""
        self.__color = WHITE
        self.__opponent = ""
        self.__silence = False
        self.__titles = None

        self.__ratings = {}

    def getName(self):
        """Returns the name of the user, without any title sufixes"""
        return self.__name

    def getStatus(self):
        """Returns the current user-status as a STATUS constant"""
        return self.__status

    def getUpTime(self):
        """Returns the when the user logged on
        Not set when status == STATUS_OFFLINE"""
        return self.__upTime + time() - self.__fingerTime

    def getIdleTime(self):
        """Returns the when the last time the user did something active
        Not set when status == STATUS_OFFLINE"""
        return self.__idleTime + time() - self.__fingerTime

    def getBusyMessage(self):
        """Returns the userset idle message
        This is set when status == STATUS_BUSY or sometimes when status ==
        STATUS_PLAYING"""
        return self.__busyMessage

    def getLastSeen(self):
        """Returns when the user logged off
        This is only set when status == STATUS_OFFLINE
        This is not set, if the user has never logged on"""
        return self.__lastSeen

    def getTotalTimeOnline(self):
        """Returns how many seconds the user has been on FICS since the account
        was created.
        This is not set, if the user has never logged on"""
        return self.__totalTimeOnline

    def getCreated(self):
        """Returns when the account was created"""
        return self.__created

    def getEmail(self):
        """Returns the email adress of the user.
        This will probably only be set for the logged in user"""
        return self.__email

    def getSanctions(self):
        """Returns any sanctions the user has against them. This is usually
        an empty string"""
        return self.__sanctions

    def getAdminLevel(self):
        """Returns the admin level as a string
        Only set for admins."""
        return self.__adminLevel

    def getTimeseal(self):
        """Returns True if the user is using timeseal for fics connection"""
        return self.__timeseal

    def getNotes(self):
        """Returns a list of the ten finger notes"""
        return self.__notes

    def getGameno(self):
        """Returns the gameno of the game in which user is currently playing
        This is only set when status == STATUS_PLAYING"""
        return self.__gameno

    def getColor(self):
        """If status == STATUS_PLAYING getColor returns the color witch the
        player has got in the game.
        Otherwise always WHITE is returned"""
        return self.__color

    def getOpponent(self):
        """Returns the opponent of the user in his current game
        This is only set when status == STATUS_PLAYING"""
        return self.__opponent

    def getSilence(self):
        """Return True if the user is playing in silence
        This is only set when status == STATUS_PLAYING"""
        return self.__silence

    def getRatings(self):
        return self.__ratings

    def getRating(self, type=None):
        return int(self.__ratings[type][0])

    def getRatingsLen(self):
        return len(self.__ratings)

    def getTitles(self):
        return self.__titles

    def setName(self, value):
        self.__name = value

    def setStatus(self, value):
        self.__status = value

    def setUpTime(self, value):
        """Use relative seconds"""
        self.__upTime = value

    def setIdleTime(self, value):
        """Use relative seconds"""
        self.__idleTime = value

    def setBusyMessage(self, value):
        """Use relative seconds"""
        self.__busyMessage = value

    def setLastSeen(self, value):
        """Use relative seconds"""
        self.__lastSeen = value

    def setTotalTimeOnline(self, value):
        """Use relative seconds"""
        self.__totalTimeOnline = value

    def setCreated(self, value):
        """Use relative seconds"""
        self.__created = value

    def setEmail(self, value):
        self.__email = value

    def setSanctions(self, value):
        self.__sanctions = value

    def setAdminLevel(self, value):
        self.__adminLevel = value

    def setTimeseal(self, value):
        self.__timeseal = value

    def setNote(self, index, note):
        self.__notes[index] = note

    def setGameno(self, value):
        self.__gameno = value

    def setColor(self, value):
        self.__color = value

    def setOpponent(self, value):
        self.__opponent = value

    def setSilence(self, value):
        self.__silence = value

    def setRating(self, rating_type, rating_line):
        self.__ratings[rating_type] = rating_line

    def setTitles(self, titles):
        self.__titles = titles


class FingerManager(GObject.GObject):
    __gsignals__ = {
        "fingeringFinished": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "ratingAdjusted": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        fingerLines = (
            r"(?P<never>%s has never connected\.)" % names,
            "Last disconnected: (?P<last>.+)",
            "On for: (?P<uptime>.+?) +Idle: (?P<idletime>.+)",
            r"%s is in (?P<silence>silence) mode\." % names,
            r"\(playing game (?P<gameno>\d+): (?P<p1>\S+?)%s vs. (?P<p2>\S+?)%s\)"
            % (titles, titles),
            r"\(%s (?P<busymessage>.+?)\)" % names,
            r"%s has not played any rated games\." % names,
            "rating +RD +win +loss +draw +total +best",
            "(?P<gametype>%s) +(?P<ratings>.+)" % types,
            "Email *: (?P<email>.+)",
            "Sanctions *: (?P<sanctions>.+)",
            "Total time online: (?P<tto>.+)",
            r"% of life online:  [\d\.]+  \(since (?P<created>.+?)\)",
            r"Timeseal [ \d] : (?P<timeseal>Off|On)",
            "Admin Level: (?P<adminlevel>.+)",
            r"(?P<noteno>\d+): *(?P<note>.*)",
            "$",
        )

        self.connection.expect_fromplus(
            self.onFinger, "Finger of %s:" % names, "$|".join(fingerLines)
        )

        self.connection.client.run_command("iset nowrap 1")

        # We don't use this. Rather we use BoardManagers "gameEnded", after
        # which we do a refinger. This is to ensure not only rating, but also
        # wins/looses/draws are updated
        # self.connection.expect(self.onRatingAdjust,
        #        "%s rating adjustment: (\d+) --> (\d+)" % types
        # Notice if you uncomment this: The expression has to be compiled with
        # re.IGNORECASE, or the first letter of 'type' must be capital

    def parseDate(self, date):
        # Tue Mar 11, 10:56 PDT 2008
        return date

    def parseShortDate(self, date):
        # 24-Oct-2007
        return date

    def parseTime(self, time):
        # 3 days, 2 hrs, 53 mins
        return time

    def onFinger(self, matchlist):
        finger = FingerObject()
        name = matchlist[0].groups()[0]
        finger.setName(name)
        if matchlist[0].groups()[1]:
            titles = re.findall(titleslist, matchlist[0].groups()[1])
            finger.setTitles(titles)
        for match in matchlist[1:]:
            if not match.group():
                continue
            groupdict = match.groupdict()
            if groupdict["never"] is not None:
                finger.setStatus(IC_STATUS_OFFLINE)
            elif groupdict["last"] is not None:
                finger.setStatus(IC_STATUS_OFFLINE)
                finger.setLastSeen(self.parseDate(groupdict["last"]))
            elif groupdict["uptime"] is not None:
                finger.setStatus(IC_STATUS_ACTIVE)
                finger.setUpTime(self.parseTime(groupdict["uptime"]))
                finger.setIdleTime(self.parseTime(groupdict["idletime"]))
            elif groupdict["silence"] is not None:
                finger.setSilence(True)
            elif groupdict["gameno"] is not None:
                finger.setStatus(IC_STATUS_PLAYING)
                finger.setGameno(groupdict["gameno"])
                if groupdict["p1"].lower() == self.connection.getUsername().lower():
                    finger.setColor(WHITE)
                    finger.setOpponent(groupdict["p2"])
                else:
                    finger.setColor(BLACK)
                    finger.setOpponent(groupdict["p1"])
            elif groupdict["busymessage"] is not None:
                finger.setStatus(IC_STATUS_BUSY)
                finger.setBusyMessage(groupdict["busymessage"])
            elif groupdict["gametype"] is not None:
                gametype = GAME_TYPES_BY_FICS_NAME[groupdict["gametype"].lower()]
                ratings = groupdict["ratings"].split()
                finger.setRating(gametype.rating_type, ratings)
            elif groupdict["email"] is not None:
                finger.setEmail(groupdict["email"])
            elif groupdict["sanctions"] is not None:
                finger.setSanctions(groupdict["sanctions"])
            elif groupdict["tto"] is not None:
                finger.setTotalTimeOnline(self.parseTime(groupdict["tto"]))
            elif groupdict["created"] is not None:
                finger.setCreated(self.parseDate(groupdict["created"]))
            elif groupdict["timeseal"] is not None:
                finger.setTimeseal(groupdict["timeseal"] == "On")
            elif groupdict["adminlevel"] is not None:
                finger.setAdminLevel(groupdict["adminlevel"])
            elif groupdict["noteno"] is not None:
                finger.setNote(int(groupdict["noteno"]) - 1, groupdict["note"])
            else:
                log.debug("Ignored fingerline: %s" % repr(match.group()))

        self.emit("fingeringFinished", finger)

    onFinger.BLKCMD = BLKCMD_FINGER

    def onRatingAdjust(self, match):
        # Notice: This is only recived for us, not for other persons we finger
        rating_type, old, new = match.groups()
        self.emit("ratingAdjusted", rating_type, new)

    def finger(self, user):
        self.connection.client.run_command("finger %s" % user)

    def setFingerNote(self, note, message):
        assert 1 <= note <= 10
        self.connection.client.run_command("set %d %s" % (note, message))

    def setBusyMessage(self, message):
        """Like set busy is really busy right now."""
        self.connection.client.run_command("set busy %s" % message)



"""This module contains chess logic functins for the pychess client. They are
based upon the lutils modules, but supports standard object types and is
therefore not as fast."""

from .lutils import lmovegen
from .lutils.validator import validateMove
from .lutils.lmove import FCORD, TCORD
from .lutils import ldraw
from .Cord import Cord
from .Move import Move
from .const import (
    LOSERSCHESS,
    WHITE,
    WHITEWON,
    BLACKWON,
    WON_NOMATERIAL,
    KING,
    HORDECHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    ATOMICCHESS,
    WON_KINGEXPLODE,
    KINGOFTHEHILLCHESS,
    BLACK,
    DRAW,
    CRAZYHOUSECHESS,
    WON_KINGINCENTER,
    THREECHECKCHESS,
    WON_THREECHECK,
    WON_MATE,
    DRAW_STALEMATE,
    DRAW_INSUFFICIENT,
    DRAW_EQUALMATERIAL,
    WON_LESSMATERIAL,
    WON_WIPEOUT,
    DRAW_REPETITION,
    WON_KINGINEIGHTROW,
    RACINGKINGSCHESS,
    DRAW_50MOVES,
    DRAW_KINGSINEIGHTROW,
    RUNNING,
    ENPASSANT,
    UNKNOWN_REASON,
)

from .lutils.bitboard import iterBits
from .lutils.attack import getAttacks
from pychess.Variants.suicide import pieceCount
from pychess.Variants.losers import testKingOnly
from pychess.Variants.atomic import kingExplode
from pychess.Variants.kingofthehill import testKingInCenter
from pychess.Variants.threecheck import checkCount
from pychess.Variants.racingkings import testKingInEightRow, test2KingInEightRow


def getDestinationCords(board, cord):
    tcords = []
    for move in lmovegen.genAllMoves(board.board):
        if FCORD(move) == cord.cord:
            if not board.board.willLeaveInCheck(move):
                tcords.append(Cord(TCORD(move)))
    return tcords


def isClaimableDraw(board):
    lboard = board.board
    if lboard.repetitionCount() >= 3:
        return True
    if ldraw.testFifty(lboard):
        return True
    return False


def playerHasMatingMaterial(board, playercolor):
    if board.variant == CRAZYHOUSECHESS:
        return True
    lboard = board.board
    return ldraw.testPlayerMatingMaterial(lboard, playercolor)


def getStatus(board):
    lboard = board.board

    if board.variant == LOSERSCHESS:
        if testKingOnly(lboard):
            if board.color == WHITE:
                status = WHITEWON
            else:
                status = BLACKWON
            return status, WON_NOMATERIAL
    elif board.variant == SUICIDECHESS or board.variant == GIVEAWAYCHESS:
        if pieceCount(lboard, lboard.color) == 0:
            if board.color == WHITE:
                status = WHITEWON
            else:
                status = BLACKWON
            return status, WON_NOMATERIAL
    elif board.variant == HORDECHESS:
        if pieceCount(lboard, lboard.color) == 0 and board.color == WHITE:
            status = BLACKWON
            return status, WON_WIPEOUT
    elif board.variant == ATOMICCHESS:
        if lboard.boards[board.color][KING] == 0:
            if board.color == WHITE:
                status = BLACKWON
            else:
                status = WHITEWON
            return status, WON_KINGEXPLODE
    elif board.variant == KINGOFTHEHILLCHESS:
        if testKingInCenter(lboard):
            if board.color == BLACK:
                status = WHITEWON
            else:
                status = BLACKWON
            return status, WON_KINGINCENTER
    elif board.variant == THREECHECKCHESS:
        if checkCount(lboard, lboard.color) == 3:
            if board.color == BLACK:
                status = WHITEWON
            else:
                status = BLACKWON
            return status, WON_THREECHECK
    elif board.variant == RACINGKINGSCHESS:
        if test2KingInEightRow(lboard):
            return DRAW, DRAW_KINGSINEIGHTROW
        elif testKingInEightRow(lboard):
            can_save = False
            for move in lmovegen.genAllMoves(lboard):
                if lboard.willGiveCheck(move) or lboard.willLeaveInCheck(move):
                    continue

                lboard.applyMove(move)
                if testKingInEightRow(lboard):
                    can_save = True
                    lboard.popMove()
                    break
                lboard.popMove()
            if not can_save:
                if board.color == BLACK:
                    status = WHITEWON
                else:
                    status = BLACKWON
                return status, WON_KINGINEIGHTROW
    else:
        if ldraw.testMaterial(lboard):
            return DRAW, DRAW_INSUFFICIENT

    hasMove = False
    for move in lmovegen.genAllMoves(lboard):
        if board.variant == ATOMICCHESS:
            if kingExplode(lboard, move, 1 - board.color) and not kingExplode(
                lboard, move, board.color
            ):
                hasMove = True
                break
            elif kingExplode(lboard, move, board.color):
                continue
        lboard.applyMove(move)
        if lboard.opIsChecked():
            lboard.popMove()
            continue
        hasMove = True
        lboard.popMove()
        break

    if not hasMove:
        if lboard.isChecked():
            if board.variant == LOSERSCHESS:
                if board.color == WHITE:
                    status = WHITEWON
                else:
                    status = BLACKWON
            else:
                if board.color == WHITE:
                    status = BLACKWON
                else:
                    status = WHITEWON
            return status, WON_MATE
        else:
            if board.variant == LOSERSCHESS or board.variant == GIVEAWAYCHESS:
                if board.color == WHITE:
                    status = WHITEWON
                else:
                    status = BLACKWON
                return status, DRAW_STALEMATE
            elif board.variant == SUICIDECHESS:
                if pieceCount(lboard, WHITE) == pieceCount(lboard, BLACK):
                    return status, DRAW_EQUALMATERIAL
                else:
                    if board.color == WHITE and pieceCount(lboard, WHITE) < pieceCount(
                        lboard, BLACK
                    ):
                        status = WHITEWON
                    else:
                        status = BLACKWON
                    return status, WON_LESSMATERIAL
            else:
                return DRAW, DRAW_STALEMATE

    if lboard.repetitionCount() >= 3:
        return DRAW, DRAW_REPETITION

    if ldraw.testFifty(lboard):
        return DRAW, DRAW_50MOVES

    return RUNNING, UNKNOWN_REASON


def standard_validate(board, move):
    return validateMove(board.board, move.move) and not board.board.willLeaveInCheck(
        move.move
    )


def validate(board, move):
    if board.variant == LOSERSCHESS:
        capture = move.flag == ENPASSANT or board[move.cord1] is not None
        if capture:
            return standard_validate(board, move)
        else:
            can_capture = False
            can_escape_with_capture = False
            ischecked = board.board.isChecked()
            for c in lmovegen.genCaptures(board.board):
                if board.board.willLeaveInCheck(c):
                    continue
                else:
                    can_capture = True
                    if ischecked:
                        can_escape_with_capture = True
                    break
            if can_capture:
                if ischecked and not can_escape_with_capture:
                    return standard_validate(board, move)
                else:
                    return False
            else:
                return standard_validate(board, move)
    elif board.variant == SUICIDECHESS:
        capture = move.flag == ENPASSANT or board[move.cord1] is not None
        if capture:
            return standard_validate(board, move)
        else:
            can_capture = False
            for c in lmovegen.genCaptures(board.board):
                can_capture = True
            if can_capture:
                return False
            else:
                return standard_validate(board, move)
    elif board.variant == ATOMICCHESS:
        # Moves exploding our king are not allowed
        if kingExplode(board.board, move.move, board.color):
            return False
        # Exploding oppont king takes precedence over mate
        elif kingExplode(board.board, move.move, 1 - board.color) and validateMove(
            board.board, move.move
        ):
            return True
        else:
            return standard_validate(board, move)
    elif board.variant == RACINGKINGSCHESS:
        # Giving check is forbidden
        if board.board.willGiveCheck(move.move):
            return False
        else:
            return standard_validate(board, move)
    else:
        return standard_validate(board, move)


def getMoveKillingKing(board):
    """Returns a move from the current color, able to capture the opponent
    king"""

    lboard = board.board
    color = lboard.color
    opking = lboard.kings[1 - color]

    for cord in iterBits(getAttacks(lboard, opking, color)):
        return Move(Cord(cord), Cord(opking), board)


def genCastles(board):
    for move in lmovegen.genCastles(board.board):
        yield Move(move)


def legalMoveCount(board):
    moves = 0
    for move in lmovegen.genAllMoves(board.board):
        if not board.board.willLeaveInCheck(move):
            moves += 1
    return moves



"""Theban Chess Variant"""

from pychess.Utils.const import THEBANCHESS, VARIANTS_OTHER
from pychess.Utils.Board import Board

THEBANSTART = "1p6/2p3kn/3p2pp/4pppp/5ppp/8/PPPPPPPP/PPPPPPKN w - - 0 1"


class ThebanBoard(Board):
    variant = THEBANCHESS
    __desc__ = _(
        "Variant developed by Kai Laskos: http://talkchess.com/forum/viewtopic.php?t=40990"
    )
    name = _("Theban")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_OTHER

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=THEBANSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



# Corner Chess

import random

from pychess.Utils.const import VARIANTS_SHUFFLE, CORNERCHESS
from pychess.Utils.Board import Board


class CornerBoard(Board):
    variant = CORNERCHESS
    __desc__ = _(
        "http://brainking.com/en/GameRules?tp=2\n"
        + "* Placement of the pieces on the 1st and 8th row are randomized\n"
        + "* The king is in the right hand corner\n"
        + "* Bishops must start on opposite color squares\n"
        + "* Black's starting position is obtained by rotating white's position 180 degrees around the board's center\n"
        + "* No castling"
    )
    name = _("Corner")
    cecp_name = "nocastle"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.shuffle_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def shuffle_start(self):
        set1 = set2 = 0
        back_rank = ["r", "n", "b", "q", "b", "n", "r"]
        while set1 % 2 == set2 % 2:
            random.shuffle(back_rank)
            set1 = back_rank.index("b")
            set2 = back_rank.index("b", set1 + 1)
        fen = "".join(back_rank)
        fen = (
            "k"
            + fen
            + "/pppppppp/8/8/8/8/PPPPPPPP/"
            + fen[::-1].upper()
            + "K w - - 0 1"
        )

        return fen


if __name__ == "__main__":
    Board = CornerBoard(True)
    for i in range(10):
        print(Board.shuffle_start())



"""Standard Chess game"""

from pychess.Utils.Board import Board


class NormalBoard(Board):
    """:Description: A standard chess game board setup"""

    __desc__ = _("Classic chess rules\n" + "http://en.wikipedia.org/wiki/Chess")
    name = _("Normal")
    cecp_name = "normal"
    need_initial_board = False
    standard_rules = True



"""Pawns Pushed Chess"""

from pychess.Utils.const import PAWNSPUSHEDCHESS, VARIANTS_OTHER
from pychess.Utils.Board import Board

PAWNSPUSHEDSTART = "rnbqkbnr/8/8/pppppppp/PPPPPPPP/8/8/RNBQKBNR w - - 0 1"


class PawnsPushedBoard(Board):
    """:Description: Standard chess rules but the start setup position is all the white pawns
    start on the 4th rank and all the black pawns start on the 5th rank
    """

    variant = PAWNSPUSHEDCHESS
    __desc__ = _(
        "FICS wild/8: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "Pawns start on 4th and 5th ranks rather than 2nd and 7th"
    )
    name = _("Pawns Pushed")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_OTHER

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=PAWNSPUSHEDSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import sys

from .attack import staticExchangeEvaluate
from .ldata import PIECE_VALUES, ASEAN_PIECE_VALUES, PAWN_VALUE, MATE_VALUE
from .lmove import GATE_PIECE
from pychess.Utils.const import (
    DROP,
    EMPTY,
    ASEAN_VARIANTS,
    PROMOTIONS,
    ATOMICCHESS,
    GATINGS,
)
from pychess.Utils.eval import pos as position_values
from pychess.Variants.atomic import kingExplode


def getCaptureValue(board, move):
    if board.variant in ASEAN_VARIANTS:
        mpV = ASEAN_PIECE_VALUES[board.arBoard[move >> 6 & 63]]
        cpV = ASEAN_PIECE_VALUES[board.arBoard[move & 63]]
    else:
        mpV = PIECE_VALUES[board.arBoard[move >> 6 & 63]]
        cpV = PIECE_VALUES[board.arBoard[move & 63]]
    if mpV < cpV:
        return cpV - mpV
    else:
        temp = staticExchangeEvaluate(board, move)
        return temp < 0 and -sys.maxsize or temp


def sortCaptures(board, moves):
    def sort_captures_func(move):
        return getCaptureValue(board, move)

    moves.sort(key=sort_captures_func, reverse=True)
    return moves


def getMoveValue(board, table, depth, move):
    """Sort criteria is as follows.
    1.  The move from the hash table
    2.  Captures as above.
    3.  Killers.
    4.  History.
    5.  Moves to the centre."""

    # As we only return directly from transposition table if hashf == hashfEXACT
    # There could be a non  hashfEXACT very promising move for us to test

    if table.isHashMove(depth, move):
        return sys.maxsize

    fcord = (move >> 6) & 63
    tcord = move & 63
    flag = move >> 12

    arBoard = board.arBoard
    fpiece = fcord if flag == DROP else arBoard[fcord]
    tpiece = arBoard[tcord]

    if tpiece != EMPTY:
        if board.variant == ATOMICCHESS:
            if kingExplode(board, move, board.color):
                return MATE_VALUE
        # We add some extra to ensure also bad captures will be searched early
        if board.variant in ASEAN_VARIANTS:
            return ASEAN_PIECE_VALUES[tpiece] - PIECE_VALUES[fpiece] + 1000
        else:
            return PIECE_VALUES[tpiece] - PIECE_VALUES[fpiece] + 1000

    if flag in PROMOTIONS:
        if board.variant in ASEAN_VARIANTS:
            return ASEAN_PIECE_VALUES[flag - 3] - PAWN_VALUE + 1000
        else:
            return PIECE_VALUES[flag - 3] - PAWN_VALUE + 1000

    if flag == DROP:
        return PIECE_VALUES[tpiece] + 1000

    if flag in GATINGS:
        return PIECE_VALUES[GATE_PIECE(flag)] + 1000

    killervalue = table.isKiller(depth, move)
    if killervalue:
        return 1000 + killervalue

    # King tropism - a move that brings us nearer to the enemy king, is probably
    # a good move
    # opking = board.kings[1-board.color]
    # score = distance[fpiece][fcord][opking] - distance[fpiece][tcord][opking]

    if board.variant in ASEAN_VARIANTS:
        score = 0
    else:
        try:
            score = (
                position_values[fpiece][board.color][tcord]
                - position_values[fpiece][board.color][fcord]
            )
            # print("NOT EMPTY fpiece", fpiece, fcord, tcord)
            # print(board)
        except KeyError:
            print("EMPTY fpiece!!!", fpiece, fcord, tcord)
            print(board)
            raise

    # History heuristic
    score += table.getButterfly(move)

    return score


def sortMoves(board, table, ply, hashmove, moves):
    def sort_moves_func(move):
        return getMoveValue(board, table, ply, hashmove, move)

    moves.sort(key=sort_moves_func, reverse=True)
    return moves



from time import time

from gi.repository import GLib, GObject

from pychess.Utils.const import WHITE, BLACK
from pychess.System.Log import log


class TimeModel(GObject.GObject):
    __gsignals__ = {
        "player_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "time_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "zero_reached": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "pause_changed": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    ############################################################################
    # Initing                                                                  #
    ############################################################################

    def __init__(self, secs=0, gain=0, bsecs=-1, minutes=-1, moves=0):
        GObject.GObject.__init__(self)
        if bsecs < 0:
            bsecs = secs
        if minutes < 0:
            minutes = secs / 60
        self.minutes = minutes  # The number of minutes for the original starting
        self.moves = moves

        # time control (not necessarily where the game was resumed,
        # i.e. self.intervals[0][0])
        if secs == 0 and gain > 0:
            self.intervals = [[gain], [gain]]
        else:
            self.intervals = [[secs], [bsecs]]
        self.gain = gain
        self.secs = secs

        # to know if game is played on ICS or not
        self.gamemodel = None

        # in FICS games we don't count gain
        self.handle_gain = True

        self.paused = False
        # The left number of secconds at the time pause was turned on
        self.pauseInterval = 0
        self.counter = None

        self.started = False
        self.ended = False

        self.movingColor = WHITE

        self.connect("time_changed", self.__zerolistener, "time_changed")
        self.connect("player_changed", self.__zerolistener, "player_changed")
        self.connect("pause_changed", self.__zerolistener, "pause_changed")

        self.zero_listener_id = None
        self.zero_listener_time = 0
        self.zero_listener_source = None

    def __repr__(self):
        text = "<TimeModel object at {} (White: {} Black: {} ended={})>".format(
            id(self),
            str(self.getPlayerTime(WHITE)),
            str(self.getPlayerTime(BLACK)),
            self.ended,
        )
        return text

    def __zerolistener(self, *args):
        if self.ended:
            return False

        cur_time = time()
        whites_time = cur_time + self.getPlayerTime(WHITE)
        blacks_time = cur_time + self.getPlayerTime(BLACK)
        if self.movingColor == WHITE:
            the_time = whites_time
            color = WHITE
        else:
            the_time = blacks_time
            color = BLACK

        remaining_time = the_time - cur_time + 0.01
        if remaining_time > 0 and remaining_time != self.zero_listener_time:
            if (
                (self.zero_listener_id is not None)
                and (self.zero_listener_source is not None)
                and not self.zero_listener_source.is_destroyed()
            ):
                GLib.source_remove(self.zero_listener_id)
            self.zero_listener_time = remaining_time
            self.zero_listener_id = GLib.timeout_add(10, self.__checkzero, color)
            default_context = (
                GLib.main_context_get_thread_default() or GLib.main_context_default()
            )
            if hasattr(default_context, "find_source_by_id"):
                self.zero_listener_source = default_context.find_source_by_id(
                    self.zero_listener_id
                )

    def __checkzero(self, color):
        if self.getPlayerTime(color) <= 0 and self.started:
            self.emit("time_changed")
            self.emit("zero_reached", color)
            return False
        return True

    ############################################################################
    # Interacting                                                              #
    ############################################################################

    def setMovingColor(self, movingColor):
        self.movingColor = movingColor
        self.emit("player_changed")

    def tap(self):
        if self.paused:
            return

        gain = self.gain if self.handle_gain else 0
        ticker = self.intervals[self.movingColor][-1] + gain
        if self.started:
            if self.counter is not None:
                ticker -= time() - self.counter
        else:
            # FICS rule
            if self.gamemodel.isPlayingICSGame():
                if self.ply >= 1:
                    self.started = True
            else:
                self.started = True
        if self.moves == 0:
            self.intervals[self.movingColor].append(ticker)
        else:
            if len(self.intervals[self.movingColor]) % self.moves == 0:
                self.intervals[self.movingColor].append(
                    self.intervals[self.movingColor][0]
                )
            else:
                self.intervals[self.movingColor].append(ticker)

        self.movingColor = 1 - self.movingColor

        if self.started:
            self.counter = time()
            self.emit("time_changed")

        self.emit("player_changed")

    def start(self):
        if self.started:
            return
        self.counter = time()
        self.emit("time_changed")

    def end(self):
        log.debug("TimeModel.end: self=%s" % self)
        self.pause()
        self.ended = True
        if (
            (self.zero_listener_id is not None)
            and (self.zero_listener_source is not None)
            and not self.zero_listener_source.is_destroyed()
        ):
            GLib.source_remove(self.zero_listener_id)

    def pause(self):
        log.debug("TimeModel.pause: self=%s" % self)
        if self.paused:
            return
        self.paused = True

        if self.counter is not None:
            self.pauseInterval = time() - self.counter

        self.counter = None
        self.emit("time_changed")
        self.emit("pause_changed", True)

    def resume(self):
        log.debug("TimeModel.resume: self=%s" % self)
        if not self.paused:
            return

        self.paused = False
        self.counter = time() - self.pauseInterval

        self.emit("pause_changed", False)

    ############################################################################
    # Undo and redo in TimeModel                                               #
    ############################################################################

    def undoMoves(self, moves):
        """Sets time and color to move, to the values they were having in the
            beginning of the ply before the current.
        his move.
        Example:
        White intervals (is thinking): [120, 130, ...]
        Black intervals:               [120, 115]
        Is undoed to:
        White intervals:               [120, 130]
        Black intervals (is thinking): [120, ...]"""

        if not self.started:
            self.start()

        for move in range(moves):
            self.movingColor = 1 - self.movingColor
            del self.intervals[self.movingColor][-1]

        if len(self.intervals[0]) + len(self.intervals[1]) >= 4:
            self.counter = time()
        else:
            self.started = False
            self.counter = None

        self.emit("time_changed")
        self.emit("player_changed")

    ############################################################################
    # Updating                                                                 #
    ############################################################################

    def updatePlayer(self, color, secs):
        self.intervals[color][-1] = secs
        if color == self.movingColor and self.started:
            self.counter = secs + time() - self.intervals[color][-1]
        self.emit("time_changed")

    ############################################################################
    # Info                                                                     #
    ############################################################################

    def getPlayerTime(self, color, movecount=-1):
        if color == self.movingColor and self.started and movecount == -1:
            if self.paused:
                return max(0, self.intervals[color][movecount] - self.pauseInterval)
            elif self.counter:
                return max(
                    0, self.intervals[color][movecount] - (time() - self.counter)
                )
        return max(0, self.intervals[color][movecount])

    def getInitialTime(self):
        return self.intervals[WHITE][0]

    def getElapsedMoveTime(self, ply):
        movecount, color = divmod(ply + 1, 2)
        gain = self.gain if ply > 2 else 0
        if len(self.intervals[color]) > movecount:
            return (
                self.intervals[color][movecount - 1]
                - self.intervals[color][movecount]
                + gain
            )
        else:
            return 0

    @property
    def display_text(self):
        text = ("%d " % self.minutes) + _("min")
        if self.gain != 0:
            text += (" + %d " % self.gain) + _("sec")
        return text

    @property
    def hasTimes(self):
        return len(self.intervals[0]) > 1

    @property
    def ply(self):
        return len(self.intervals[BLACK]) + len(self.intervals[WHITE]) - 2

    def hasBWTimes(self, bmovecount, wmovecount):
        return (
            len(self.intervals[BLACK]) > bmovecount
            and len(self.intervals[WHITE]) > wmovecount
        )

    def isBlitzFide(self):
        val = 60 * self.minutes + 60 * (self.gain if self.handle_gain else 0)
        return (
            val > 0 and val <= 600
        )  # Less or equal than 10 minutes for 60 moves and for each player



from pychess.Utils.const import DROP
from pychess.Utils.lutils.lmovegen import newMove

MAXMOVE = newMove(63, 63, DROP)
COMMENT, VARI_START, VARI_END, NAG = (MAXMOVE + i + 1 for i in range(4))


def walk(node, arr, txt):
    """Prepares a game data for databse.
    Recursively walks the node tree to collect moves and comments.
    Arguments:
    node - list (a tree of lboards created by the pgn parser)
    arr - array("H") (2 byte unsigned ints representing lmove objects
                     or COMMENT, VARI_START, VARI_END, NAG+nag)
    txt - list (comment strings)"""

    arr_append = arr.append
    while True:
        if node is None:
            break

        # Initial game or variation comment
        if node.prev is None:
            for child in node.children:
                if isinstance(child, str):
                    arr_append(COMMENT)
                    txt.append(child)
            node = node.next
            continue

        arr_append(node.lastMove)

        for nag in node.nags:
            if nag:
                arr_append(NAG + int(nag[1:]))

        for child in node.children:
            if isinstance(child, str):
                # comment
                arr_append(COMMENT)
                txt.append(child)
            else:
                # variations
                arr_append(VARI_START)
                walk(child[0], arr, txt)
                arr_append(VARI_END)

        if node.next:
            node = node.next
        else:
            break



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from pychess.Utils.IconLoader import get_pixbuf, load_icon

# from pychess.widgets.WebKitBrowser import open_link

main_window = None

gtk_close = load_icon(16, "gtk-close", "window-close")


def mainwindow():
    return main_window


def createImage(pixbuf):
    image = Gtk.Image()
    image.set_from_pixbuf(pixbuf)
    return image


def createAlignment(top, right, bottom, left):
    align = Gtk.Alignment.new(0.5, 0.5, 1, 1)
    align.set_property("top-padding", top)
    align.set_property("right-padding", right)
    align.set_property("bottom-padding", bottom)
    align.set_property("left-padding", left)
    return align


def new_notebook(name=None):
    def customGetTabLabelText(child):
        return name

    notebook = Gtk.Notebook()
    if name is not None:
        notebook.set_name(name)
    notebook.get_tab_label_text = customGetTabLabelText
    notebook.set_show_tabs(False)
    notebook.set_show_border(False)
    return notebook


def dock_panel_tab(title, desc, icon, button=None):
    box = Gtk.Box()
    pixbuf = get_pixbuf(icon, 16)
    image = Gtk.Image.new_from_pixbuf(pixbuf)
    label = Gtk.Label(label=title)
    box.set_tooltip_text(desc)
    box.pack_start(image, False, True, 0)
    box.pack_start(label, False, True, 0)
    if button is not None:
        box.pack_start(button, False, True, 0)

    box.set_spacing(2)
    box.show_all()

    return box


def insert_formatted(text_view, iter, text, tag=None):
    def insert(text):
        if tag is not None:
            tb.insert_with_tags_by_name(iter, text, tag)
        else:
            tb.insert(iter, text)

    tb = text_view.get_buffer()
    # I know this is far from perfect but I don't want to use re for this
    if "://" in text or "www" in text:
        parts = text.split()
        position = 0
        for i, part in enumerate(parts):
            if "://" in part or "www" in part:
                if part.startswith('"'):
                    part = part[1:]
                    endpos = part.find('"')
                    if endpos != -1:
                        part = part[:endpos]
                part0 = (
                    "http://web.archive.org/%s" % part
                    if part.startswith("http://www.endgame.nl")
                    else part
                )
                parts[i] = f'<a href="{part0}">{part}</a>'
                position = i
                break
        insert("%s " % " ".join(parts[:position]))
        label = Gtk.Label()
        label.set_markup(parts[position])
        # label.connect("activate-link", open_link)
        label.show()
        anchor = tb.create_child_anchor(iter)
        text_view.add_child_at_anchor(label, anchor)
        insert(" %s" % " ".join(parts[position + 1 :]))
    else:
        insert(text)



import asyncio
import random
from io import StringIO

from gi.repository import Gtk

from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import (
    WHITE,
    BLACK,
    LOCAL,
    NORMALCHESS,
    ARTIFICIAL,
    chr2Sign,
    chrU2Sign,
    FAN_PIECES,
    HINT,
    ENDGAME,
)
from pychess.Utils.LearnModel import LearnModel
from pychess.Utils.TimeModel import TimeModel
from pychess.Utils.lutils.attack import isAttacked
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import FILE, RANK
from pychess.Variants import variants
from pychess.Players.Human import Human
from pychess.Players.engineNest import discoverer
from pychess.System import conf
from pychess.perspectives import perspective_manager
from pychess.Savers import fen as fen_loader

__title__ = _("Endgames")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _("Practice endgames with computer")


# TODO: get it from a text file
ENDGAMES = (
    ("kpk", "King and Pawn vs King"),
    ("kbnk", "King, Bishop and Knight vs King"),
    ("kbbk", "King and 2 Bishops vs King"),
    ("krk", "King and Rook vs King"),
    ("kqk", "King and Queen vs King"),
    ("kqkr", "King and Queen vs King and Rook"),
    ("krpkr", "King, Rook and Pawn vs King and Rook"),
    ("kppkp", "King and 2 Pawns vs King and Pawn"),
    ("kpkp", "King and Pawn vs King and Pawn"),
    ("kqpkq", "King, Queen and Pawn vs King and Queen"),
    ("knnkp", "King and Two Knights and vs King and Pawn"),
    ("kppkpp", "King and two pawns vs King and two pawns"),
    ("kqqkqr", "King and two queens vs King and Queen"),
)


class Sidepanel:
    def load(self, persp):
        self.persp = persp
        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.tv = Gtk.TreeView()

        renderer = Gtk.CellRendererText()
        renderer.props.font = "Times 14"
        column = Gtk.TreeViewColumn(_("White"), renderer, text=0)
        self.tv.append_column(column)

        renderer = Gtk.CellRendererText()
        renderer.props.font = "Times 14"
        column = Gtk.TreeViewColumn(_("Black"), renderer, text=1)
        self.tv.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Title"), renderer, text=2)
        self.tv.append_column(column)

        self.tv.connect("row-activated", self.row_activated)

        self.store = Gtk.ListStore(str, str, str)

        for pieces, title in ENDGAMES:
            if pieces.count("k") != 2:
                print("Game needs exactly 2 kings! %s" % pieces)
                continue
            elif len(pieces) > 6:
                print("Max 6 pieces, please! %s" % pieces)
                continue
            else:
                for piece in pieces:
                    if piece not in ("kqrbnp"):
                        print(f"Invalid piece {piece} in {pieces}")
                        continue

                pos = pieces.rfind("k")
                white_pieces, black_pieces = pieces[:pos], pieces[pos:]
                wfan = []
                for piece in white_pieces:
                    wfan.append(FAN_PIECES[0][chr2Sign[piece]])
                bfan = []
                for piece in black_pieces:
                    bfan.append(FAN_PIECES[1][chr2Sign[piece]])
                self.store.append(["".join(wfan), "".join(bfan), title])

        self.tv.set_model(self.store)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.BROWSE)
        self.tv.set_cursor(conf.get("learncombo%s" % ENDGAME))

        scrollwin = Gtk.ScrolledWindow()
        scrollwin.add(self.tv)
        scrollwin.show_all()

        self.box.pack_start(scrollwin, True, True, 0)
        self.box.show_all()

        return self.box

    def row_activated(self, widget, path, col):
        if path is None:
            return
        else:
            pieces = ENDGAMES[path[0]][0].lower()
            conf.set("categorycombo", ENDGAME)
            from pychess.widgets.TaskerManager import learn_tasker

            learn_tasker.learn_combo.set_active(path[0])
            start_endgame_from(pieces)


def start_endgame_from(pieces):
    fen = create_fen(pieces)

    timemodel = TimeModel(0, 0)
    gamemodel = LearnModel(timemodel)
    gamemodel.set_learn_data(ENDGAME, pieces)

    player_name = conf.get("firstName")
    p0 = (LOCAL, Human, (WHITE, player_name), player_name)

    engine = discoverer.getEngineByName(discoverer.getEngineLearn())
    ponder_off = True
    engine_name = discoverer.getName(engine)
    p1 = (
        ARTIFICIAL,
        discoverer.initPlayerEngine,
        (engine, BLACK, 20, variants[NORMALCHESS], 60, 0, 0, ponder_off),
        engine_name,
    )

    def restart_analyzer(gamemodel):
        asyncio.create_task(gamemodel.restart_analyzer(HINT))

    gamemodel.connect("learn_success", restart_analyzer)

    def on_game_started(gamemodel):
        perspective.activate_panel("annotationPanel")
        asyncio.create_task(
            gamemodel.start_analyzer(HINT, force_engine=discoverer.getEngineLearn())
        )

    gamemodel.connect("game_started", on_game_started)

    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(
        perspective.generalStart(
            gamemodel, p0, p1, loaddata=(StringIO(fen), fen_loader, 0, -1)
        )
    )


def create_fen(pieces):
    """Create a random FEN position using given pieces"""

    pos = pieces.rfind("k")
    pieces = pieces[:pos], pieces[pos:]

    ok = False
    while not ok:
        lboard = LBoard()
        lboard.applyFen("8/8/8/8/8/8/8/8 w - - 0 1")
        bishop_cords = [[], []]
        bishop_colors_ok = True

        cords = list(range(0, 64))
        pawn_cords = list(range(0 + 8, 64 - 8))

        # Order of color is important here to prevent offering
        # positions with trivial captures in first move
        for color in (WHITE, BLACK):
            for char in pieces[color]:
                piece = chrU2Sign[char.upper()]
                attacked = True
                limit = 100
                while attacked and limit > 0:
                    cord = random.choice(pawn_cords if char == "p" else cords)
                    attacked = isAttacked(lboard, cord, 1 - color)
                    limit -= 1
                lboard._addPiece(cord, piece, color)
                cords.remove(cord)
                if cord in pawn_cords:
                    pawn_cords.remove(cord)
                if char == "b":
                    bishop_cords[color].append(cord)

            # 2 same color bishop is not ok
            if len(bishop_cords[color]) == 2 and bishop_colors_ok:
                b0, b1 = bishop_cords[color]
                b0_color = BLACK if RANK(b0) % 2 == FILE(b0) % 2 else WHITE
                b1_color = BLACK if RANK(b1) % 2 == FILE(b1) % 2 else WHITE
                if b0_color == b1_color:
                    bishop_colors_ok = False
                    break

        ok = (
            (not lboard.isChecked()) and (not lboard.opIsChecked()) and bishop_colors_ok
        )

    fen = lboard.asFen()
    return fen



from gi.repository import GObject

from pychess.System.Log import log
from pychess.Utils.const import UNSUPPORTED
from pychess.ic import GAME_TYPES, TITLES, RATING_TYPES
from pychess.ic.FICSObjects import FICSSeek
from pychess.ic.icc import DG_SEEK, DG_SEEK_REMOVED
from pychess.ic.managers.SeekManager import SeekManager


class ICCSeekManager(SeekManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(DG_SEEK, self.on_icc_seek_add)
        self.connection.expect_dg_line(DG_SEEK_REMOVED, self.on_icc_seek_removed)

        self.connection.client.run_command("set-2 %s 1" % DG_SEEK)
        self.connection.client.run_command("set-2 %s 1" % DG_SEEK_REMOVED)

    def on_icc_seek_add(self, data):
        log.debug("DG_SEEK_ADD %s" % data)
        # index name titles rating provisional-status wild rating-type time
        # inc rated color minrating maxrating autoaccept formula fancy-time-control
        # 195 Tinker {C} 2402 2 0 Blitz 5 3 1 -1 0 9999 1 1 {}

        parts = data.split(" ", 2)
        index = int(parts[0])
        player = self.connection.players.get(parts[1])

        titles_end = parts[2].find("}")
        titles = parts[2][1:titles_end]
        tit = set()
        for title in titles.split():
            tit.add(TITLES[title])
        player.titles |= tit

        parts = parts[2][titles_end + 1 :].split()
        rating = int(parts[0])
        deviation = None  # parts[1]
        # wild = parts[2]
        try:
            gametype = GAME_TYPES[parts[3].lower()]
        except KeyError:
            return
        minutes = int(parts[4])
        increment = int(parts[5])
        rated = parts[6] == "1"
        color = parts[7]
        if color == "-1":
            color = None
        else:
            color = "white" if color == "1" else "black"
        rmin = int(parts[8])
        rmax = int(parts[9])
        automatic = parts[10] == "1"
        # formula = parts[11]
        # fancy_tc = parts[12]

        if gametype.variant_type in UNSUPPORTED:
            log.debug("!!! unsupported variant in seek: %s" % data)
            return

        if (
            gametype.rating_type in RATING_TYPES
            and player.ratings[gametype.rating_type] != rating
        ):
            player.ratings[gametype.rating_type] = rating
            player.deviations[gametype.rating_type] = deviation
            player.emit("ratings_changed", gametype.rating_type, player)

        seek = FICSSeek(
            index,
            player,
            minutes,
            increment,
            rated,
            color,
            gametype,
            rmin=rmin,
            rmax=rmax,
            automatic=automatic,
        )
        self.emit("addSeek", seek)

    def on_icc_seek_removed(self, data):
        log.debug("DG_SEEK_REMOVED %s" % data)
        key = data.split()[0]
        self.emit("removeSeek", int(key))
